"""
`rgba`:
    An RGB/RGBA color: is a tuple of 3 integers, representing the red (`0`-`255`), green (`0`-`255`), and blue (`0`-`255`).
    Also includes an optional 4th param, which is a float, that represents the alpha channel (`0.0`-`1.0`).
`hsla`:
    An HSL/HSB color: is a tuple of 3 integers, representing the hue (`0`-`360`), saturation (`0`-`100`), and lightness (`0`-`100`).
    Also includes an optional 4th param, which is a float, that represents the alpha channel (`0.0`-`1.0`).
`hexa`:
    A HEX color: is a string in the format `RGB`, `RGBA`, `RRGGBB` or `RRGGBBAA` (where `R` `G` `B` `A` are hexadecimal digits).

-------------------------------------------------------------------------------------------------------------------------------------
The `Color` class, which contains all sorts of different color-related methods:
- validate colors:
    - is valid rgba
    - is valid hsla
    - is valid hexa
    - is valid in any format
- check if a color has an alpha channel
- convert between different color formats:
    - color to rgba
    - color to hsla
    - color to hexa
- recognize colors inside strings and convert them to color types:
    - string to rgba
- convert an RGBA color to a HEX integer
- convert a HEX integer to an RGBA color
- get a colors luminance from the RGB channels
- get the optimal text color for on a colored background
- adjust different color channels:
    - brightness
    - saturation
"""

from .xx_regex import Regex

import re as _re


class rgba:
    """An RGB/RGBA color: is a tuple of 3 integers, representing the red (`0`-`255`), green (`0`-`255`), and blue (`0`-`255`).\n
    Also includes an optional 4th param, which is a float, that represents the alpha channel (`0.0`-`1.0`).\n
    -----------------------------------------------------------------------------------------------------------------------------
    Includes methods:
    - `to_hsla()` to convert to HSL color
    - `to_hexa()` to convert to HEX color
    - `has_alpha()` to check if the color has an alpha channel
    - `lighten(amount)` to create a lighter version of the color
    - `darken(amount)` to create a darker version of the color
    - `saturate(amount)` to increase color saturation
    - `desaturate(amount)` to decrease color saturation
    - `rotate(degrees)` to rotate the hue by degrees
    - `invert()` to get the inverse color
    - `grayscale()` to convert to grayscale
    - `blend(other, ratio)` to blend with another color
    - `is_dark()` to check if the color is considered dark
    - `is_light()` to check if the color is considered light
    - `is_grayscale()` to check if the color is grayscale
    - `is_opaque()` to check if the color has no transparency
    - `with_alpha(alpha)` to create a new color with different alpha
    - `complementary()` to get the complementary color"""

    def __init__(self, r: int, g: int, b: int, a: float = None, _validate: bool = True):
        if not _validate:
            self.r, self.g, self.b, self.a = r, g, b, a
            return
        if any(isinstance(x, rgba) for x in (r, g, b)):
            raise ValueError("Color is already a rgba() color")
        elif not all(isinstance(x, int) and 0 <= x <= 255 for x in (r, g, b)):
            raise ValueError(
                "RGBA color must have R G B as integers in [0, 255]: got",
                (r, g, b),
            )
        elif a is not None and not (isinstance(a, (int, float)) and 0 <= a <= 1):
            raise ValueError(f"Alpha channel must be a float/int in [0.0, 1.0]: got '{a}'")
        self.r, self.g, self.b = r, g, b
        self.a = None if a is None else (1.0 if a > 1.0 else float(a))

    def __len__(self) -> int:
        return 3 if self.a is None else 4

    def __iter__(self) -> iter:
        return iter((self.r, self.g, self.b) + (() if self.a is None else (self.a,)))

    def __dict__(self) -> dict:
        return self.dict()

    def __getitem__(self, index: int) -> int:
        return ((self.r, self.g, self.b) + (() if self.a is None else (self.a,)))[index]

    def __repr__(self) -> str:
        return f'rgba({self.r}, {self.g}, {self.b}{"" if self.a is None else f", {self.a}"})'

    def __str__(self) -> str:
        return f'({self.r}, {self.g}, {self.b}{"" if self.a is None else f", {self.a}"})'

    def __eq__(self, other: "rgba") -> bool:
        if not isinstance(other, rgba):
            return False
        return (self.r, self.g, self.b, self.a) == (
            other[0],
            other[1],
            other[2],
            other[3],
        )

    def dict(self) -> dict:
        """Returns the color components as a dictionary with keys `'r'`, `'g'`, `'b'` and optionally `'a'`"""
        return dict(r=self.r, g=self.g, b=self.b) if self.a is None else dict(r=self.r, g=self.g, b=self.b, a=self.a)

    def values(self) -> tuple:
        """Returns the color components as separate values `r, g, b, a`"""
        return self.r, self.g, self.b, self.a

    def to_hsla(self) -> "hsla":
        """Returns the color as a `hsla()` color"""
        return hsla(*self._rgb_to_hsl(self.r, self.g, self.b), self.a, _validate=False)

    def to_hexa(self) -> "hexa":
        """Returns the color as a `hexa()` color"""
        return hexa("", self.r, self.g, self.b, self.a)

    def has_alpha(self) -> bool:
        """Returns `True` if the color has an alpha channel and `False` otherwise"""
        return self.a is not None

    def lighten(self, amount: float) -> "rgba":
        """Increases the colors lightness by the specified amount (`0.0`-`1.0`)"""
        self.r, self.g, self.b, self.a = self.to_hsla().lighten(amount).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def darken(self, amount: float) -> "rgba":
        """Decreases the colors lightness by the specified amount (`0.0`-`1.0`)"""
        self.r, self.g, self.b, self.a = self.to_hsla().darken(amount).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def saturate(self, amount: float) -> "rgba":
        """Increases the colors saturation by the specified amount (`0.0`-`1.0`)"""
        self.r, self.g, self.b, self.a = self.to_hsla().saturate(amount).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def desaturate(self, amount: float) -> "rgba":
        """Decreases the colors saturation by the specified amount (`0.0`-`1.0`)"""
        self.r, self.g, self.b, self.a = self.to_hsla().desaturate(amount).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def rotate(self, degrees: int) -> "rgba":
        """Rotates the colors hue by the specified number of degrees"""
        self.r, self.g, self.b, self.a = self.to_hsla().rotate(degrees).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def invert(self, invert_alpha: bool = False) -> "rgba":
        """Inverts the color by rotating hue by 180 degrees and inverting lightness"""
        self.r, self.g, self.b = 255 - self.r, 255 - self.g, 255 - self.b
        if invert_alpha:
            self.a = 1 - self.a
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def grayscale(self) -> "rgba":
        """Converts the color to grayscale using the luminance formula"""
        self.r = self.g = self.b = Color.luminance(self.r, self.g, self.b)
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def blend(self, other: "rgba", ratio: float = 0.5, additive_alpha: bool = False) -> "rgba":
        """Blends the current color with another color using the specified ratio (`0.0`-`1.0`):
        - if `ratio` is `0.0` it means 100% of the current color and 0% of the `other` color (2:0 mixture)
        - if `ratio` is `0.5` it means 50% of both colors (1:1 mixture)
        - if `ratio` is `1.0` it means 0% of the current color and 100% of the `other` color (0:2 mixture)"""
        if not (isinstance(ratio, (int, float)) and 0 <= ratio <= 1):
            raise ValueError("'ratio' must be a float/int in [0.0, 1.0]")
        elif not isinstance(other, rgba):
            if Color.is_valid_rgba(other):
                other = rgba(*other, _validate=False)
            else:
                raise TypeError("'other' must be a valid RGBA color")
        ratio *= 2
        self.r = max(0, min(255, int(round((self.r * (2 - ratio)) + (other.r * ratio)))))
        self.g = max(0, min(255, int(round((self.g * (2 - ratio)) + (other.g * ratio)))))
        self.b = max(0, min(255, int(round((self.b * (2 - ratio)) + (other.b * ratio)))))
        none_alpha = self.a is None and (len(other) <= 3 or other[3] is None)
        if not none_alpha:
            self_a = 1 if self.a is None else self.a
            other_a = (other[3] if other[3] is not None else 1) if len(other) > 3 else 1
            if additive_alpha:
                self.a = max(0, min(1, (self_a * (2 - ratio)) + (other_a * ratio)))
            else:
                self.a = max(
                    0,
                    min(
                        1,
                        (self_a * (1 - (ratio / 2))) + (other_a * (ratio / 2)),
                    ),
                )
        else:
            self.a = None
        return rgba(self.r, self.g, self.b, None if none_alpha else self.a, _validate=False)

    def is_dark(self) -> bool:
        """Returns `True` if the color is considered dark (`lightness < 50%`)"""
        return self.to_hsla().is_dark()

    def is_light(self) -> bool:
        """Returns `True` if the color is considered light (`lightness >= 50%`)"""
        return not self.is_dark()

    def is_grayscale(self) -> bool:
        """Returns `True` if the color is grayscale"""
        return self.r == self.g == self.b

    def is_opaque(self) -> bool:
        """Returns `True` if the color has no transparency"""
        return self.a == 1 or self.a is None

    def with_alpha(self, alpha: float) -> "rgba":
        """Returns a new color with the specified alpha value"""
        if not (isinstance(alpha, (int, float)) and 0 <= alpha <= 1):
            raise ValueError("'alpha' must be a float/int in [0.0, 1.0]")
        return rgba(self.r, self.g, self.b, alpha, _validate=False)

    def complementary(self) -> "rgba":
        """Returns the complementary color (180 degrees on the color wheel)"""
        return self.to_hsla().complementary().to_rgba()

    def _rgb_to_hsl(self, r: int, g: int, b: int) -> tuple:
        r, g, b = r / 255.0, g / 255.0, b / 255.0
        max_c, min_c = max(r, g, b), min(r, g, b)
        l = (max_c + min_c) / 2
        if max_c == min_c:
            h = s = 0
        else:
            delta = max_c - min_c
            s = delta / (1 - abs(2 * l - 1))
            if max_c == r:
                h = ((g - b) / delta) % 6
            elif max_c == g:
                h = ((b - r) / delta) + 2
            else:
                h = ((r - g) / delta) + 4
            h /= 6
        return int(round(h * 360)), int(round(s * 100)), int(round(l * 100))


class hsla:
    """A HSL/HSLA color: is a tuple of 3 integers, representing hue (`0`-`360`), saturation (`0`-`100`), and lightness (`0`-`100`).\n
    Also includes an optional 4th param, which is a float, that represents the alpha channel (`0.0`-`1.0`).\n
    ----------------------------------------------------------------------------------------------------------------------------------
    Includes methods:
    - `to_rgba()` to convert to RGB color
    - `to_hexa()` to convert to HEX color
    - `has_alpha()` to check if the color has an alpha channel
    - `lighten(amount)` to create a lighter version of the color
    - `darken(amount)` to create a darker version of the color
    - `saturate(amount)` to increase color saturation
    - `desaturate(amount)` to decrease color saturation
    - `rotate(degrees)` to rotate the hue by degrees
    - `invert()` to get the inverse color
    - `grayscale()` to convert to grayscale
    - `blend(other, ratio)` to blend with another color
    - `is_dark()` to check if the color is considered dark
    - `is_light()` to check if the color is considered light
    - `is_grayscale()` to check if the color is grayscale
    - `is_opaque()` to check if the color has no transparency
    - `with_alpha(alpha)` to create a new color with different alpha
    - `complementary()` to get the complementary color"""

    def __init__(self, h: int, s: int, l: int, a: float = None, _validate: bool = True):
        if not _validate:
            self.h, self.s, self.l, self.a = h, s, l, a
            return
        if any(isinstance(x, hsla) for x in (h, s, l)):
            raise ValueError("Color is already a hsla() color")
        elif not (isinstance(h, int) and (0 <= h <= 360) and all(isinstance(x, int) and (0 <= x <= 100) for x in (s, l))):
            raise ValueError(
                "HSL color must have H as integer in [0, 360] and S L as integers in [0, 100]: got",
                (h, s, l),
            )
        elif a is not None and (not isinstance(a, (int, float)) or not 0 <= a <= 1):
            raise ValueError(f"Alpha channel must be a float/int in [0.0, 1.0]: got '{a}'")
        self.h, self.s, self.l = h, s, l
        self.a = None if a is None else (1.0 if a > 1.0 else float(a))

    def __len__(self) -> int:
        return 3 if self.a is None else 4

    def __iter__(self) -> iter:
        return iter((self.h, self.s, self.l) + (() if self.a is None else (self.a,)))

    def __dict__(self) -> dict:
        return self.dict()

    def __getitem__(self, index: int) -> int:
        return ((self.h, self.s, self.l) + (() if self.a is None else (self.a,)))[index]

    def __repr__(self) -> str:
        return f'hsla({self.h}, {self.s}, {self.l}{"" if self.a is None else f", {self.a}"})'

    def __str__(self) -> str:
        return f'({self.h}, {self.s}, {self.l}{"" if self.a is None else f", {self.a}"})'

    def __eq__(self, other: "hsla") -> bool:
        if not isinstance(other, hsla):
            return False
        return (self.h, self.s, self.l, self.a) == (
            other[0],
            other[1],
            other[2],
            other[3],
        )

    def dict(self) -> dict:
        """Returns the color components as a dictionary with keys `'h'`, `'s'`, `'l'` and optionally `'a'`"""
        return dict(h=self.h, s=self.s, l=self.l) if self.a is None else dict(h=self.h, s=self.s, l=self.l, a=self.a)

    def values(self) -> tuple:
        """Returns the color components as separate values `h, s, l, a`"""
        return self.h, self.s, self.l, self.a

    def to_rgba(self) -> "rgba":
        """Returns the color as a `rgba()` color"""
        return rgba(*self._hsl_to_rgb(self.h, self.s, self.l), self.a, _validate=False)

    def to_hexa(self) -> "hexa":
        """Returns the color as a `hexa()` color"""
        r, g, b = self._hsl_to_rgb(self.h, self.s, self.l)
        return hexa("", r, g, b, self.a)

    def has_alpha(self) -> bool:
        """Returns `True` if the color has an alpha channel and `False` otherwise"""
        return self.a is not None

    def lighten(self, amount: float) -> "hsla":
        """Increases the colors lightness by the specified amount (`0.0`-`1.0`)"""
        if not (isinstance(amount, (int, float)) and 0 <= amount <= 1):
            raise ValueError("'amount' must be a float/int in [0.0, 1.0]")
        self.l = int(min(100, self.l + (100 - self.l) * amount))
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def darken(self, amount: float) -> "hsla":
        """Decreases the colors lightness by the specified amount (`0.0`-`1.0`)"""
        if not (isinstance(amount, (int, float)) and 0 <= amount <= 1):
            raise ValueError("'amount' must be a float/int in [0.0, 1.0]")
        self.l = int(max(0, self.l * (1 - amount)))
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def saturate(self, amount: float) -> "hsla":
        """Increases the colors saturation by the specified amount (`0.0`-`1.0`)"""
        if not (isinstance(amount, (int, float)) and 0 <= amount <= 1):
            raise ValueError("'amount' must be a float/int in [0.0, 1.0]")
        self.s = int(min(100, self.s + (100 - self.s) * amount))
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def desaturate(self, amount: float) -> "hsla":
        """Decreases the colors saturation by the specified amount (`0.0`-`1.0`)"""
        if not (isinstance(amount, (int, float)) and 0 <= amount <= 1):
            raise ValueError("'amount' must be a float/int in [0.0, 1.0]")
        self.s = int(max(0, self.s * (1 - amount)))
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def rotate(self, degrees: int) -> "hsla":
        """Rotates the colors hue by the specified number of degrees"""
        self.h = (self.h + degrees) % 360
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def invert(self, invert_alpha: bool = False) -> "hsla":
        """Inverts the color by rotating hue by 180 degrees and inverting lightness"""
        self.h = (self.h + 180) % 360
        self.l = 100 - self.l
        if invert_alpha:
            self.a = 1 - self.a
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def grayscale(self) -> "hsla":
        """Converts the color to grayscale using the luminance formula"""
        l = Color.luminance(*self._hsl_to_rgb(self.h, self.s, self.l))
        self.h, self.s, self.l, _ = rgba(l, l, l, _validate=False).to_hsla().values()
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def blend(self, other: "hsla", ratio: float = 0.5, additive_alpha: bool = False) -> "rgba":
        """Blends the current color with another color using the specified ratio (`0.0`-`1.0`):
        - if `ratio` is `0.0` it means 100% of the current color and 0% of the `other` color (2:0 mixture)
        - if `ratio` is `0.5` it means 50% of both colors (1:1 mixture)
        - if `ratio` is `1.0` it means 0% of the current color and 100% of the `other` color (0:2 mixture)"""
        self.h, self.s, self.l, self.a = self.to_rgba().blend(Color.to_rgba(other), ratio, additive_alpha).to_hsla().values()
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def is_dark(self) -> bool:
        """Returns `True` if the color is considered dark (`lightness < 50%`)"""
        return self.l < 50

    def is_light(self) -> bool:
        """Returns `True` if the color is considered light (`lightness >= 50%`)"""
        return not self.is_dark()

    def is_grayscale(self) -> bool:
        """Returns `True` if the color is considered grayscale"""
        return self.s == 0

    def is_opaque(self) -> bool:
        """Returns `True` if the color has no transparency"""
        return self.a == 1 or self.a is None

    def with_alpha(self, alpha: float) -> "hsla":
        """Returns a new color with the specified alpha value"""
        if not (isinstance(alpha, (int, float)) and 0 <= alpha <= 1):
            raise ValueError("'alpha' must be a float/int in [0.0, 1.0]")
        return hsla(self.h, self.s, self.l, alpha, _validate=False)

    def complementary(self) -> "hsla":
        """Returns the complementary color (180 degrees on the color wheel)"""
        return hsla((self.h + 180) % 360, self.s, self.l, self.a, _validate=False)

    def _hsl_to_rgb(self, h: int, s: int, l: int) -> tuple:
        h, s, l = h / 360, s / 100, l / 100
        if s == 0:
            r = g = b = int(l * 255)
        else:

            def hue_to_rgb(p, q, t):
                if t < 0:
                    t += 1
                if t > 1:
                    t -= 1
                if t < 1 / 6:
                    return p + (q - p) * 6 * t
                if t < 1 / 2:
                    return q
                if t < 2 / 3:
                    return p + (q - p) * (2 / 3 - t) * 6
                return p

            q = l * (1 + s) if l < 0.5 else l + s - l * s
            p = 2 * l - q
            r = int(round(hue_to_rgb(p, q, h + 1 / 3) * 255))
            g = int(round(hue_to_rgb(p, q, h) * 255))
            b = int(round(hue_to_rgb(p, q, h - 1 / 3) * 255))
        return r, g, b


class hexa:
    """A HEX color: is a string representing a hexadecimal color code with optional alpha channel.\n
    -------------------------------------------------------------------------------------------------
    Supports formats: RGB, RGBA, RRGGBB, RRGGBBAA (with or without prefix)
    Includes methods:
    - `to_rgba()` to convert to RGB color
    - `to_hsla()` to convert to HSL color
    - `has_alpha()` to check if the color has an alpha channel
    - `lighten(amount)` to create a lighter version of the color
    - `darken(amount)` to create a darker version of the color
    - `saturate(amount)` to increase color saturation
    - `desaturate(amount)` to decrease color saturation
    - `rotate(degrees)` to rotate the hue by degrees
    - `invert()` to get the inverse color
    - `grayscale()` to convert to grayscale
    - `blend(other, ratio)` to blend with another color
    - `is_dark()` to check if the color is considered dark
    - `is_light()` to check if the color is considered light
    - `is_grayscale()` to check if the color is grayscale
    - `is_opaque()` to check if the color has no transparency
    - `with_alpha(alpha)` to create a new color with different alpha
    - `complementary()` to get the complementary color"""

    def __init__(self, color: str | int, _r: int = None, _g: int = None, _b: int = None, _a: float = None):
        if all(x is not None for x in (_r, _g, _b)):
            self.r, self.g, self.b, self.a = _r, _g, _b, _a
            return
        if isinstance(color, hexa):
            raise ValueError("Color is already a hexa() color")
        if isinstance(color, str):
            if color.startswith("#"):
                color = color[1:].upper()
            elif color.startswith("0x"):
                color = color[2:].upper()
            if len(color) == 3:  # RGB
                self.r, self.g, self.b, self.a = (
                    int(color[0] * 2, 16),
                    int(color[1] * 2, 16),
                    int(color[2] * 2, 16),
                    None,
                )
            elif len(color) == 4:  # RGBA
                self.r, self.g, self.b, self.a = (
                    int(color[0] * 2, 16),
                    int(color[1] * 2, 16),
                    int(color[2] * 2, 16),
                    int(color[3] * 2, 16) / 255.0,
                )
            elif len(color) == 6:  # RRGGBB
                self.r, self.g, self.b, self.a = (
                    int(color[0:2], 16),
                    int(color[2:4], 16),
                    int(color[4:6], 16),
                    None,
                )
            elif len(color) == 8:  # RRGGBBAA
                self.r, self.g, self.b, self.a = (
                    int(color[0:2], 16),
                    int(color[2:4], 16),
                    int(color[4:6], 16),
                    int(color[6:8], 16) / 255.0,
                )
            else:
                raise ValueError(f"Invalid HEX format '{color}'")
        elif isinstance(color, int):
            self.r, self.g, self.b, self.a = Color.hex_int_to_rgba(color).values()
        else:
            raise TypeError(f"HEX color must be of type 'str' or 'int': got '{type(color)}'")

    def __len__(self) -> int:
        return 3 if self.a is None else 4

    def __iter__(self) -> iter:
        return iter(
            (f"{self.r:02X}", f"{self.g:02X}", f"{self.b:02X}") + (() if self.a is None else (f"{int(self.a * 255):02X}",))
        )

    def __dict__(self) -> dict:
        return self.dict()

    def __getitem__(self, index: int) -> int:
        return (
            (f"{self.r:02X}", f"{self.g:02X}", f"{self.b:02X}") + (() if self.a is None else (f"{int(self.a * 255):02X}",))
        )[index]

    def __repr__(self) -> str:
        return f'hexa(#{self.r:02X}{self.g:02X}{self.b:02X}{"" if self.a is None else f"{int(self.a * 255):02X}"})'

    def __str__(self) -> str:
        return f'#{self.r:02X}{self.g:02X}{self.b:02X}{"" if self.a is None else f"{int(self.a * 255):02X}"}'

    def __eq__(self, other: "hexa") -> bool:
        if not isinstance(other, hexa):
            return False
        return (self.r, self.g, self.b, self.a) == (
            other[0],
            other[1],
            other[2],
            other[3],
        )

    def dict(self) -> dict:
        """Returns the color components as a dictionary with hex string values for keys `'r'`, `'g'`, `'b'` and optionally `'a'`"""
        return (
            dict(r=f"{self.r:02X}", g=f"{self.g:02X}", b=f"{self.b:02X}")
            if self.a is None
            else dict(
                r=f"{self.r:02X}",
                g=f"{self.g:02X}",
                b=f"{self.b:02X}",
                a=f"{int(self.a * 255):02X}",
            )
        )

    def values(self) -> tuple:
        """Returns the color components as separate values `r, g, b, a`"""
        return self.r, self.g, self.b, self.a

    def to_rgba(self, round_alpha: bool = True) -> "rgba":
        """Returns the color as a `rgba()` color"""
        return rgba(
            self.r,
            self.g,
            self.b,
            None if self.a is None else (round(self.a, 2) if round_alpha else self.a),
            _validate=False,
        )

    def to_hsla(self, round_alpha: bool = True) -> "hsla":
        """Returns the color as a `hsla()` color"""
        return self.to_rgba(round_alpha).to_hsla()

    def has_alpha(self) -> bool:
        """Returns `True` if the color has an alpha channel and `False` otherwise"""
        return self.a is not None

    def lighten(self, amount: float) -> "hexa":
        """Increases the colors lightness by the specified amount (`0.0`-`1.0`)"""
        self.r, self.g, self.b, self.a = self.to_rgba(False).lighten(amount).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def darken(self, amount: float) -> "hexa":
        """Decreases the colors lightness by the specified amount (`0.0`-`1.0`)"""
        self.r, self.g, self.b, self.a = self.to_rgba(False).darken(amount).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def saturate(self, amount: float) -> "hexa":
        """Increases the colors saturation by the specified amount (`0.0`-`1.0`)"""
        self.r, self.g, self.b, self.a = self.to_rgba(False).saturate(amount).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def desaturate(self, amount: float) -> "hexa":
        """Decreases the colors saturation by the specified amount (`0.0`-`1.0`)"""
        self.r, self.g, self.b, self.a = self.to_rgba(False).desaturate(amount).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def rotate(self, degrees: int) -> "hexa":
        """Rotates the colors hue by the specified number of degrees"""
        self.r, self.g, self.b, self.a = self.to_rgba(False).rotate(degrees).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def invert(self, invert_alpha: bool = False) -> "hexa":
        """Inverts the color by rotating hue by 180 degrees and inverting lightness"""
        self.r, self.g, self.b, self.a = self.to_rgba(False).invert().values()
        if invert_alpha:
            self.a = 1 - self.a
        return hexa("", self.r, self.g, self.b, self.a)

    def grayscale(self) -> "hexa":
        """Converts the color to grayscale using the luminance formula"""
        self.r = self.g = self.b = Color.luminance(self.r, self.g, self.b)
        return hexa("", self.r, self.g, self.b, self.a)

    def blend(self, other: "hexa", ratio: float = 0.5, additive_alpha: bool = False) -> "rgba":
        """Blends the current color with another color using the specified ratio (`0.0`-`1.0`):
        - if `ratio` is `0.0` it means 100% of the current color and 0% of the `other` color (2:0 mixture)
        - if `ratio` is `0.5` it means 50% of both colors (1:1 mixture)
        - if `ratio` is `1.0` it means 0% of the current color and 100% of the `other` color (0:2 mixture)"""
        self.r, self.g, self.b, self.a = self.to_rgba(False).blend(Color.to_rgba(other), ratio, additive_alpha).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def is_dark(self) -> bool:
        """Returns `True` if the color is considered dark (`lightness < 50%`)"""
        return self.to_hsla(False).is_dark()

    def is_light(self) -> bool:
        """Returns `True` if the color is considered light (`lightness >= 50%`)"""
        return not self.is_dark()

    def is_grayscale(self) -> bool:
        """Returns `True` if the color is grayscale (`saturation == 0`)"""
        return self.to_hsla(False).is_grayscale()

    def is_opaque(self) -> bool:
        """Returns `True` if the color has no transparency (`alpha == 1.0`)"""
        return self.a == 1 or self.a is None

    def with_alpha(self, alpha: float) -> "hexa":
        """Returns a new color with the specified alpha value"""
        if not (isinstance(alpha, (int, float)) and 0 <= alpha <= 1):
            raise ValueError("'alpha' must be in [0.0, 1.0]")
        return hexa("", self.r, self.g, self.b, alpha)

    def complementary(self) -> "hexa":
        """Returns the complementary color (180 degrees on the color wheel)"""
        return self.to_hsla(False).complementary().to_hexa()


class Color:

    @staticmethod
    def is_valid_rgba(color: str | list | tuple | dict, allow_alpha: bool = True) -> bool:
        try:
            if isinstance(color, rgba):
                return True
            elif isinstance(color, (list, tuple)):
                if allow_alpha and Color.has_alpha(color):
                    return (
                        0 <= color[0] <= 255
                        and 0 <= color[1] <= 255
                        and 0 <= color[2] <= 255
                        and (0 <= color[3] <= 1 or color[3] is None)
                    )
                return 0 <= color[0] <= 255 and 0 <= color[1] <= 255 and 0 <= color[2] <= 255
            elif isinstance(color, dict):
                if allow_alpha and Color.has_alpha(color):
                    return (
                        0 <= color["r"] <= 255
                        and 0 <= color["g"] <= 255
                        and 0 <= color["b"] <= 255
                        and (0 <= color["a"] <= 1 or color["a"] is None)
                    )
                return 0 <= color["r"] <= 255 and 0 <= color["g"] <= 255 and 0 <= color["b"] <= 255
            elif isinstance(color, str):
                return bool(_re.fullmatch(Regex.rgba_str(allow_alpha=allow_alpha), color))
            return False
        except Exception:
            return False

    @staticmethod
    def is_valid_hsla(color: str | list | tuple | dict, allow_alpha: bool = True) -> bool:
        try:
            if isinstance(color, hsla):
                return True
            elif isinstance(color, (list, tuple)):
                if allow_alpha and Color.has_alpha(color):
                    return (
                        0 <= color[0] <= 360
                        and 0 <= color[1] <= 100
                        and 0 <= color[2] <= 100
                        and (0 <= color[3] <= 1 or color[3] is None)
                    )
                else:
                    return 0 <= color[0] <= 360 and 0 <= color[1] <= 100 and 0 <= color[2] <= 100
            elif isinstance(color, dict):
                if allow_alpha and Color.has_alpha(color):
                    return (
                        0 <= color["h"] <= 360
                        and 0 <= color["s"] <= 100
                        and 0 <= color["l"] <= 100
                        and (0 <= color["a"] <= 1 or color["a"] is None)
                    )
                else:
                    return 0 <= color["h"] <= 360 and 0 <= color["s"] <= 100 and 0 <= color["l"] <= 100
            elif isinstance(color, str):
                return bool(_re.fullmatch(Regex.hsla_str(allow_alpha=allow_alpha), color))
        except Exception:
            return False

    @staticmethod
    def is_valid_hexa(color: str | int, allow_alpha: bool = True, get_prefix: bool = False) -> bool | tuple[bool, str]:
        try:
            if isinstance(color, hexa):
                return (True, "#")
            elif isinstance(color, int):
                is_valid = 0 <= color <= (0xFFFFFFFF if allow_alpha else 0xFFFFFF)
                return (is_valid, "0x") if get_prefix else is_valid
            elif isinstance(color, str):
                color, prefix = (
                    (color[1:], "#")
                    if color.startswith("#")
                    else (color[2:], "0x") if color.startswith("0x") else (color, None)
                )
                return (
                    (bool(_re.fullmatch(Regex.hexa_str(allow_alpha=allow_alpha), color)), prefix)
                    if get_prefix
                    else bool(_re.fullmatch(Regex.hexa_str(allow_alpha=allow_alpha), color))
                )
        except Exception:
            return (False, None) if get_prefix else False

    @staticmethod
    def is_valid(color: str | list | tuple | dict, allow_alpha: bool = True) -> bool:
        return (
            Color.is_valid_rgba(color, allow_alpha)
            or Color.is_valid_hsla(color, allow_alpha)
            or Color.is_valid_hexa(color, allow_alpha)
        )

    @staticmethod
    def has_alpha(color: rgba | hsla | hexa) -> bool:
        """Check if the given color has an alpha channel.\n
        ---------------------------------------------------------------------------
        Input a RGBA, HSLA or HEXA color as `color`.
        Returns `True` if the color has an alpha channel and `False` otherwise."""
        if isinstance(color, (rgba, hsla, hexa)):
            return color.has_alpha()
        if Color.is_valid_hexa(color):
            if isinstance(color, str):
                if color.startswith("#"):
                    color = color[1:]
                return len(color) == 4 or len(color) == 8
            if isinstance(color, int):
                hex_length = len(f"{color:X}")
                return hex_length == 4 or hex_length == 8
        elif isinstance(color, (list, tuple)) and len(color) == 4 and color[3] is not None:
            return True
        elif isinstance(color, dict) and len(color) == 4 and color["a"] is not None:
            return True
        return False

    @staticmethod
    def to_rgba(color: hsla | hexa) -> rgba:
        """Will try to convert any color type to a color of type RGBA."""
        if isinstance(color, (hsla, hexa)):
            return color.to_rgba()
        elif Color.is_valid_hsla(color):
            return hsla(*color, _validate=False).to_rgba()
        elif Color.is_valid_hexa(color):
            return hexa(color).to_rgba()
        elif Color.is_valid_rgba(color):
            return color if isinstance(color, rgba) else (rgba(*color, _validate=False))
        raise ValueError(f"Invalid color format '{color}'")

    @staticmethod
    def to_hsla(color: rgba | hexa) -> hsla:
        """Will try to convert any color type to a color of type HSLA."""
        if isinstance(color, (rgba, hexa)):
            return color.to_hsla()
        elif Color.is_valid_rgba(color):
            return rgba(*color, _validate=False).to_hsla()
        elif Color.is_valid_hexa(color):
            return hexa(color).to_hsla()
        elif Color.is_valid_hsla(color):
            return color if isinstance(color, hsla) else (hsla(*color, _validate=False))
        raise ValueError(f"Invalid color format '{color}'")

    @staticmethod
    def to_hexa(color: rgba | hsla) -> hexa:
        """Will try to convert any color type to a color of type HEXA."""
        if isinstance(color, (rgba, hsla)):
            return color.to_hexa()
        elif Color.is_valid_rgba(color):
            return rgba(*color, _validate=False).to_hexa()
        elif Color.is_valid_hsla(color):
            return hsla(*color, _validate=False).to_hexa()
        elif Color.is_valid_hexa(color):
            return color if isinstance(color, hexa) else hexa(color)
        raise ValueError(f"Invalid color format '{color}'")

    @staticmethod
    def str_to_rgba(string: str, only_first: bool = False) -> rgba | list[rgba] | None:
        """Will try to recognize RGBA colors inside a string and output the found ones as RGBA objects.\n
        --------------------------------------------------------------------------------------------------
        If `only_first` is `True` only the first found color will be returned (not as a list)."""
        if only_first:
            match = _re.search(Regex.rgb_str(allow_alpha=True), string)
            if not match:
                return None
            m = match.groups()
            return rgba(
                int(m[0]),
                int(m[1]),
                int(m[2]),
                ((int(m[3]) if "." not in m[3] else float(m[3])) if m[3] else None),
                _validate=False,
            )
        else:
            matches = _re.findall(Regex.rgb_str(allow_alpha=True), string)
            if not matches:
                return None
            return [
                rgba(
                    int(m[0]),
                    int(m[1]),
                    int(m[2]),
                    ((int(m[3]) if "." not in m[3] else float(m[3])) if m[3] else None),
                    _validate=False,
                )
                for m in matches
            ]

    @staticmethod
    def rgba_to_hex_int(
        r: int,
        g: int,
        b: int,
        a: float = None,
        preserve_original: bool = False,
    ) -> int:
        """Convert RGBA channels to a HEXA integer (alpha is optional).\n
        --------------------------------------------------------------------------------------------
        To preserve leading zeros, the function will add a `1` at the beginning, if the HEX integer
        would start with a `0`.
        This could affect the color a little bit, but will make sure, that it won't be interpreted
        as a completely different color, when initializing it as a `hexa()` color or changing it
        back to RGBA using `Color.hex_int_to_rgba()`.\n
        ⇾ You can disable this behavior by setting `preserve_original` to `True`"""
        r = max(0, min(255, int(r)))
        g = max(0, min(255, int(g)))
        b = max(0, min(255, int(b)))
        if a is None:
            hex_int = (r << 16) | (g << 8) | b
            if not preserve_original and (hex_int & 0xF00000) == 0:
                hex_int |= 0x010000
        else:
            a = max(0, min(255, int(a * 255)))
            hex_int = (r << 24) | (g << 16) | (b << 8) | a
            if not preserve_original and r == 0:
                hex_int |= 0x01000000
        return hex_int

    @staticmethod
    def hex_int_to_rgba(hex_int: int, preserve_original: bool = False) -> rgba:
        """Convert a HEX integer to RGBA channels.\n
        -------------------------------------------------------------------------------------------
        If the red channel is `1` after conversion, it will be set to `0`, because when converting
        from RGBA to a HEX integer, the first `0` will be set to `1` to preserve leading zeros.
        This is the correction, so the color doesn't even look slightly different.\n
        ⇾ You can disable this behavior by setting `preserve_original` to `True`"""
        if not isinstance(hex_int, int):
            raise ValueError("Input must be an integer")
        hex_str = f"{hex_int:x}"
        if len(hex_str) <= 6:
            hex_str = hex_str.zfill(6)
            return rgba(
                r if (r := int(hex_str[0:2], 16)) != 1 or preserve_original else 0,
                int(hex_str[2:4], 16),
                int(hex_str[4:6], 16),
                None,
                _validate=False,
            )
        elif len(hex_str) <= 8:
            hex_str = hex_str.zfill(8)
            return rgba(
                r if (r := int(hex_str[0:2], 16)) != 1 or preserve_original else 0,
                int(hex_str[2:4], 16),
                int(hex_str[4:6], 16),
                int(hex_str[6:8], 16) / 255.0,
                _validate=False,
            )
        else:
            raise ValueError(f"Invalid HEX integer '0x{hex_str}': expected in range [0x000000, 0xFFFFFF]")

    @staticmethod
    def luminance(r: int, g: int, b: int, output_type: type = None) -> int | float:
        """Gets the colors luminance using the luminance formula.\n
        ------------------------------------------------------------
        The param `output_type` can be set to:
        - `int`   =⠀integer in [0, 100]
        - `float` =⠀float in [0.0, 1.0]
        - `None`  =⠀integer in [0, 255]"""
        r, g, b = r / 255.0, g / 255.0, b / 255.0
        if r < 0.03928:
            r = r / 12.92
        else:
            r = ((r + 0.055) / 1.055) ** 2.4
        if g < 0.03928:
            g = g / 12.92
        else:
            g = ((g + 0.055) / 1.055) ** 2.4
        if b < 0.03928:
            b = b / 12.92
        else:
            b = ((b + 0.055) / 1.055) ** 2.4
        l = 0.2126 * r + 0.7152 * g + 0.0722 * b
        return round(l * 100) if isinstance(output_type, int) else round(l * 255) if output_type is None else l

    @staticmethod
    def text_color_for_on_bg(text_bg_color: rgba | hexa) -> rgba | hexa:
        was_hexa, was_int = Color.is_valid_hexa(text_bg_color), isinstance(text_bg_color, int)
        text_bg_color = Color.to_rgba(text_bg_color)
        brightness = 0.2126 * text_bg_color[0] + 0.7152 * text_bg_color[1] + 0.0722 * text_bg_color[2]
        return (
            (hexa("", 255, 255, 255) if was_hexa else rgba(255, 255, 255, _validate=False))
            if brightness < 128
            else ((0x000 if was_int else hexa("", 0, 0, 0)) if was_hexa else rgba(0, 0, 0, _validate=False))
        )

    @staticmethod
    def adjust_lightness(color: rgba | hexa, lightness_change: float) -> rgba | hexa:
        """In- or decrease the lightness of the input color.\n
        -----------------------------------------------------------------------------------------------------
        - color (rgba|hexa): HEX or RGBA color
        - lightness_change (float): float between -1.0 (darken by `100%`) and 1.0 (lighten by `100%`)\n
        -----------------------------------------------------------------------------------------------------
        returns (rgba|hexa): the adjusted color in the format of the input color"""
        was_hexa = Color.is_valid_hexa(color)
        color = Color.to_hsla(color)
        h, s, l, a = (
            color[0],
            color[1],
            color[2],
            color[3] if Color.has_alpha(color) else None,
        )
        l = int(max(0, min(100, l + lightness_change * 100)))
        return hsla(h, s, l, a, _validate=False).to_hexa() if was_hexa else hsla(h, s, l, a, _validate=False).to_rgba()

    @staticmethod
    def adjust_saturation(color: rgba | hexa, saturation_change: float) -> rgba | hexa:
        """In- or decrease the saturation of the input color.\n
        -----------------------------------------------------------------------------------------------------------
        - color (rgba|hexa): HEX or RGBA color
        - saturation_change (float): float between -1.0 (saturate by `100%`) and 1.0 (desaturate by `100%`)\n
        -----------------------------------------------------------------------------------------------------------
        returns (rgba|hexa): the adjusted color in the format of the input color"""
        was_hexa = Color.is_valid_hexa(color)
        color = Color.to_hsla(color)
        h, s, l, a = (
            color[0],
            color[1],
            color[2],
            color[3] if Color.has_alpha(color) else None,
        )
        s = int(max(0, min(100, s + saturation_change * 100)))
        return hsla(h, s, l, a, _validate=False).to_hexa() if was_hexa else hsla(h, s, l, a, _validate=False).to_rgba()



"""
Functions for modifying and checking the systems environment-variables:
- `EnvPath.paths()`
- `EnvPath.has_path()`
- `EnvPath.add_path()`
- `EnvPath.remove_path()`
"""

from .xx_path import Path

import sys as _sys
import os as _os


class EnvPath:

    @staticmethod
    def paths(as_list: bool = False) -> str | list:
        """Get the PATH environment variable."""
        paths = _os.environ.get("PATH", "")
        return paths.split(_os.pathsep) if as_list else paths

    @staticmethod
    def has_path(path: str = None, cwd: bool = False, base_dir: bool = False) -> bool:
        """Check if a path is present in the PATH environment variable."""
        if cwd:
            path = _os.getcwd()
        elif base_dir:
            path = Path.get(base_dir=True)
        elif path is None:
            raise ValueError("A path must be provided or either 'cwd' or 'base_dir' must be True.")
        paths = EnvPath.paths(as_list=True)
        return _os.path.normpath(path) in [_os.path.normpath(p) for p in paths]

    @staticmethod
    def add_path(
        path: str = None,
        cwd: bool = False,
        base_dir: bool = False,
    ) -> None:
        """Add a path to the PATH environment variable."""
        path = EnvPath.__get(path, cwd, base_dir)
        if not EnvPath.has_path(path):
            EnvPath.__persistent(path, add=True)

    @staticmethod
    def remove_path(
        path: str = None,
        cwd: bool = False,
        base_dir: bool = False,
    ) -> None:
        """Remove a path from the PATH environment variable."""
        path = EnvPath.__get(path, cwd, base_dir)
        if EnvPath.has_path(path):
            EnvPath.__persistent(path, remove=True)

    @staticmethod
    def __get(
        path: str = None,
        cwd: bool = False,
        base_dir: bool = False,
    ) -> list:
        """Get and/or normalize the paths.\n
        ------------------------------------------------------------------------------------
        Raise an error if no path is provided and neither `cwd` or `base_dir` is `True`."""
        if cwd:
            path = _os.getcwd()
        elif base_dir:
            path = Path.get(base_dir=True)
        elif path is None:
            raise ValueError("A path must be provided or either 'cwd' or 'base_dir' must be True.")
        return _os.path.normpath(path)

    @staticmethod
    def __persistent(path: str, add: bool = False, remove: bool = False) -> None:
        """Add or remove a path from PATH persistently across sessions as well as the current session."""
        if add == remove:
            raise ValueError("Either add or remove must be True, but not both.")
        current_paths = EnvPath.paths(as_list=True)
        path = _os.path.normpath(path)
        if remove:
            current_paths = [p for p in current_paths if _os.path.normpath(p) != _os.path.normpath(path)]
        elif add:
            current_paths.append(path)
        _os.environ["PATH"] = new_path = _os.pathsep.join(sorted(set(filter(bool, current_paths))))
        if _sys.platform == "win32":  # Windows
            try:
                import winreg as _winreg

                key = _winreg.OpenKey(
                    _winreg.HKEY_CURRENT_USER,
                    "Environment",
                    0,
                    _winreg.KEY_ALL_ACCESS,
                )
                _winreg.SetValueEx(key, "PATH", 0, _winreg.REG_EXPAND_SZ, new_path)
                _winreg.CloseKey(key)
            except ImportError:
                print("Warning: Unable to make persistent changes on Windows.")
        else:  # UNIX-like (Linux/macOS)
            shell_rc_file = _os.path.expanduser(
                "~/.bashrc" if _os.path.exists(_os.path.expanduser("~/.bashrc")) else "~/.zshrc"
            )
            with open(shell_rc_file, "r+") as f:
                content = f.read()
                f.seek(0)
                if remove:
                    new_content = [line for line in content.splitlines() if not line.endswith(f':{path}"')]
                    f.write("\n".join(new_content))
                else:
                    f.write(f'{content.rstrip()}\n# Added by XulbuX\nexport PATH="{new_path}"\n')
                f.truncate()
            _os.system(f"source {shell_rc_file}")



from xulbux import Console


def test_console_user():
    user_output = Console.user()
    assert isinstance(user_output, str)
    assert user_output != ""


def test_console_width():
    width_output = Console.w()
    assert isinstance(width_output, int)
    assert width_output > 0


def test_console_height():
    height_output = Console.h()
    assert isinstance(height_output, int)
    assert height_output > 0



from xulbux import rgba, hexa, hsla


# ! DONT'T CHANGE VALUES ! #
clr_rgba = (255, 0, 0, 0.5)
clr_hexa = "#FF00007F"
clr_hsla = (0, 100, 50, 0.5)


####################### FUNCTIONS FOR CHECKING CUSTOM TYPE VALUES #######################


def assert_rgba_equal(actual: rgba, expected: tuple):
    assert isinstance(actual, rgba)
    assert actual[0] == expected[0]
    assert actual[1] == expected[1]
    assert actual[2] == expected[2]
    assert actual[3] == expected[3]


def assert_hsla_equal(actual: hsla, expected: tuple):
    assert isinstance(actual, hsla)
    assert actual[0] == expected[0]
    assert actual[1] == expected[1]
    assert actual[2] == expected[2]
    assert actual[3] == expected[3]


def assert_hexa_equal(actual: hexa, expected: str):
    assert isinstance(actual, hexa)
    assert str(actual) == expected


######################################## TESTING ########################################


def test_rgba_return_values():
    assert_rgba_equal(rgba(*clr_rgba), clr_rgba)
    assert_hsla_equal(rgba(*clr_rgba).to_hsla(), (0, 100, 50, 0.5))
    assert_hexa_equal(rgba(*clr_rgba).to_hexa(), "#FF00007F")
    assert rgba(*clr_rgba).has_alpha() is True
    assert_rgba_equal(rgba(*clr_rgba).lighten(0.5), (255, 128, 128, 0.5))
    assert_rgba_equal(rgba(*clr_rgba).darken(0.5), (128, 0, 0, 0.5))
    assert_rgba_equal(rgba(*clr_rgba).saturate(0.5), (255, 0, 0, 0.5))
    assert_rgba_equal(rgba(*clr_rgba).desaturate(0.5), (191, 64, 64, 0.5))
    assert_rgba_equal(rgba(*clr_rgba).rotate(90), (128, 255, 0, 0.5))
    assert_rgba_equal(rgba(*clr_rgba).rotate(-90), (127, 0, 255, 0.5))
    assert_rgba_equal(rgba(*clr_rgba).invert(), (0, 255, 255, 0.5))
    assert_rgba_equal(rgba(*clr_rgba).grayscale(), (54, 54, 54, 0.5))
    assert_rgba_equal(rgba(*clr_rgba).blend((0, 255, 0)), (255, 255, 0, 0.75))
    assert rgba(*clr_rgba).is_dark() is False
    assert rgba(*clr_rgba).is_light() is True
    assert rgba(*clr_rgba).is_grayscale() is False
    assert rgba(*clr_rgba).is_opaque() is False
    assert_rgba_equal(rgba(*clr_rgba).with_alpha(0.75), (255, 0, 0, 0.75))
    assert_rgba_equal(rgba(*clr_rgba).complementary(), (0, 255, 255, 0.5))


def test_hsla_return_values():
    assert_hsla_equal(hsla(*clr_hsla), clr_hsla)
    assert_rgba_equal(hsla(*clr_hsla).to_rgba(), (255, 0, 0, 0.5))
    assert_hexa_equal(hsla(*clr_hsla).to_hexa(), "#FF00007F")
    assert hsla(*clr_hsla).has_alpha() is True
    assert_hsla_equal(hsla(*clr_hsla).lighten(0.5), (0, 100, 75, 0.5))
    assert_hsla_equal(hsla(*clr_hsla).darken(0.5), (0, 100, 25, 0.5))
    assert_hsla_equal(hsla(*clr_hsla).saturate(0.5), (0, 100, 50, 0.5))
    assert_hsla_equal(hsla(*clr_hsla).desaturate(0.5), (0, 50, 50, 0.5))
    assert_hsla_equal(hsla(*clr_hsla).rotate(90), (90, 100, 50, 0.5))
    assert_hsla_equal(hsla(*clr_hsla).rotate(-90), (270, 100, 50, 0.5))
    assert_hsla_equal(hsla(*clr_hsla).invert(), (180, 100, 50, 0.5))
    assert_hsla_equal(hsla(*clr_hsla).grayscale(), (0, 0, 21, 0.5))
    assert_hsla_equal(hsla(*clr_hsla).blend((120, 100, 50)), (60, 100, 50, 0.75))
    assert hsla(*clr_hsla).is_dark() is False
    assert hsla(*clr_hsla).is_light() is True
    assert hsla(*clr_hsla).is_grayscale() is False
    assert hsla(*clr_hsla).is_opaque() is False
    assert_hsla_equal(hsla(*clr_hsla).with_alpha(0.75), (0, 100, 50, 0.75))
    assert_hsla_equal(hsla(*clr_hsla).complementary(), (180, 100, 50, 0.5))


def test_hexa_return_values():
    assert_hexa_equal(hexa(clr_hexa), clr_hexa)
    assert_rgba_equal(hexa(clr_hexa).to_rgba(), (255, 0, 0, 0.5))
    assert_hsla_equal(hexa(clr_hexa).to_hsla(), (0, 100, 50, 0.5))
    assert hexa(clr_hexa).has_alpha() is True
    assert_hexa_equal(hexa(clr_hexa).lighten(0.5), "#FF80807F")
    assert_hexa_equal(hexa(clr_hexa).darken(0.5), "#8000007F")
    assert_hexa_equal(hexa(clr_hexa).saturate(0.5), "#FF00007F")
    assert_hexa_equal(hexa(clr_hexa).desaturate(0.5), "#BF40407F")
    assert_hexa_equal(hexa(clr_hexa).rotate(90), "#80FF007F")
    assert_hexa_equal(hexa(clr_hexa).rotate(-90), "#7F00FF7F")
    assert_hexa_equal(hexa(clr_hexa).invert(), "#00FFFF7F")
    assert_hexa_equal(hexa(clr_hexa).grayscale(), "#3636367F")
    assert_hexa_equal(hexa(clr_hexa).blend("#00FF00"), "#FFFF00BF")
    assert hexa(clr_hexa).is_dark() is False
    assert hexa(clr_hexa).is_light() is True
    assert hexa(clr_hexa).is_grayscale() is False
    assert hexa(clr_hexa).is_opaque() is False
    assert_hexa_equal(hexa(clr_hexa).with_alpha(0.75), "#FF0000BF")
    assert_hexa_equal(hexa(clr_hexa).complementary(), "#00FFFF7F")


print(rgba(*clr_rgba).is_dark())



"""
Really long regex code presets:
`quotes` match everything inside quotes
`brackets` match everything inside brackets
`outside_strings` match the pattern but not inside strings
`all_except` match everything except a certain pattern
`func_call` match a function call
`rgba_str` match an RGBA color
`hsla_str` match a HSLA color
"""

import regex as _rx
import re as _re


class Regex:

    @staticmethod
    def quotes() -> str:
        """Matches everything inside quotes. (strings)\n
        --------------------------------------------------------------------------------
        Will create two named groups:
        - `quote` the quote type (single or double)
        - `string` everything inside the found quote pair\n
        --------------------------------------------------------------------------------
        Attention: Requires non standard library `regex` not standard library `re`!"""
        return r'(?P<quote>[\'"])(?P<string>(?:\\.|(?!\g<quote>).)*?)\g<quote>'

    @staticmethod
    def brackets(bracket1: str = "(", bracket2: str = ")", is_group: bool = False, ignore_in_strings: bool = True) -> str:
        """Matches everything inside brackets, including other nested brackets.\n
        --------------------------------------------------------------------------------
        If `is_group` is true, you will be able to reference the matched content as a
        group (e.g. `match.group(…)` or `r'\\…'`).
        If `ignore_in_strings` is true and a bracket is inside a string (e.g. `'...'`
        or `"..."`), it will not be counted as the matching closing bracket.\n
        --------------------------------------------------------------------------------
        Attention: Requires non standard library `regex` not standard library `re`!"""
        g, b1, b2 = (
            "" if is_group else "?:",
            _rx.escape(bracket1) if len(bracket1) == 1 else bracket1,
            _rx.escape(bracket2) if len(bracket2) == 1 else bracket2,
        )
        if ignore_in_strings:
            return rf'{b1}\s*({g}(?:[^{b1}{b2}"\']|"(?:\\.|[^"\\])*"|\'(?:\\.|[^\'\\])*\'|{b1}(?:[^{b1}{b2}"\']|"(?:\\.|[^"\\])*"|\'(?:\\.|[^\'\\])*\'|(?R))*{b2})*)\s*{b2}'
        else:
            return rf"{b1}\s*({g}(?:[^{b1}{b2}]|{b1}(?:[^{b1}{b2}]|(?R))*{b2})*)\s*{b2}"

    @staticmethod
    def outside_strings(pattern: str = r".*") -> str:
        """Matches the `pattern` only when it is not found inside a string (`'...'` or `"..."`)."""
        return rf'(?<!["\'])(?:{pattern})(?!["\'])'

    @staticmethod
    def all_except(
        disallowed_pattern: str,
        ignore_pattern: str = "",
        is_group: bool = False,
    ) -> str:
        """Matches everything except `disallowed_pattern`, unless the `disallowed_pattern`
        is found inside a string (`'...'` or `"..."`).\n
        ------------------------------------------------------------------------------------
        The `ignore_pattern` is just always ignored. For example if `disallowed_pattern` is
        `>` and `ignore_pattern` is `->`, the `->`-arrows will be allowed, even though they
        have `>` in them.
        If `is_group` is true, you will be able to reference the matched content as a group
        (e.g. `match.group(…)` or `r'\\…'`)."""
        return rf'({"" if is_group else "?:"}(?:(?!{ignore_pattern}).)*(?:(?!{Regex.outside_strings(disallowed_pattern)}).)*)'

    @staticmethod
    def func_call(func_name: str = None) -> str:
        """Match a function call, and get back two groups:
        1. function name
        2. the function's arguments\n
        If no `func_name` is given, it will match any function call.\n
        --------------------------------------------------------------------------------
        Attention: Requires non standard library `regex` not standard library `re`!"""
        return (
            r"(?<=\b)(" + (r"[\w_]+" if func_name is None else func_name) + r")\s*" + Regex.brackets("(", ")", is_group=True)
        )

    @staticmethod
    def rgba_str(fix_sep: str = ",", allow_alpha: bool = True) -> str:
        """Matches an RGBA color inside a string.\n
        ----------------------------------------------------------------------------
        The RGBA color can be in the formats (for `fix_sep = ','`):
        - `rgba(r, g, b)`
        - `rgba(r, g, b, a)` (if `allow_alpha=True`)
        - `(r, g, b)`
        - `(r, g, b, a)` (if `allow_alpha=True`)
        - `r, g, b`
        - `r, g, b, a` (if `allow_alpha=True`)\n
        #### Valid ranges:
        - `r` 0-255 (int: red)
        - `g` 0-255 (int: green)
        - `b` 0-255 (int: blue)
        - `a` 0-1 (float: opacity)\n
        ----------------------------------------------------------------------------
        If the `fix_sep` is set to nothing, any char that is not a letter or number
        can be used to separate the RGBA values, including just a space."""
        if fix_sep in (None, ""):
            fix_sep = r"[^0-9A-Z]"
        else:
            fix_sep = _re.escape(fix_sep)
        rgb_part = rf"""((?:0*(?:25[0-5]|2[0-4][0-9]|1?[0-9]{{1,2}})))
            (?:\s*{fix_sep}\s*)((?:0*(?:25[0-5]|2[0-4][0-9]|1?[0-9]{{1,2}})))
            (?:\s*{fix_sep}\s*)((?:0*(?:25[0-5]|2[0-4][0-9]|1?[0-9]{{1,2}})))"""
        return (
            rf"""(?ix)
            (?:rgb|rgba)?\s*(?:\(?\s*{rgb_part}
                (?:(?:\s*{fix_sep}\s*)((?:0*(?:0?\.[0-9]+|1\.0+|[0-9]+\.[0-9]+|[0-9]+))))?
            \s*\)?)"""
            if allow_alpha
            else rf"(?ix)(?:rgb|rgba)?\s*(?:\(?\s*{rgb_part}\s*\)?)"
        )

    @staticmethod
    def hsla_str(fix_sep: str = ",", allow_alpha: bool = True) -> str:
        """Matches a HSLA color inside a string.\n
        ----------------------------------------------------------------------------
        The HSLA color can be in the formats (for `fix_sep = ','`):
        - `hsla(h, s, l)`
        - `hsla(h, s, l, a)` (if `allow_alpha=True`)
        - `(h, s, l)`
        - `(h, s, l, a)` (if `allow_alpha=True`)
        - `h, s, l`
        - `h, s, l, a` (if `allow_alpha=True`)\n
        #### Valid ranges:
        - `h` 0-360 (int: hue)
        - `s` 0-100 (int: saturation)
        - `l` 0-100 (int: lightness)
        - `a` 0-1 (float: opacity)\n
        ----------------------------------------------------------------------------
        If the `fix_sep` is set to nothing, any char that is not a letter or number
        can be used to separate the HSLA values, including just a space."""
        if fix_sep in (None, ""):
            fix_sep = r"[^0-9A-Z]"
        else:
            fix_sep = _re.escape(fix_sep)
        hsl_part = rf"""((?:0*(?:360|3[0-5][0-9]|[12][0-9][0-9]|[1-9]?[0-9])))
            (?:\s*{fix_sep}\s*)((?:0*(?:100|[1-9][0-9]|[0-9])))
            (?:\s*{fix_sep}\s*)((?:0*(?:100|[1-9][0-9]|[0-9])))"""
        return (
            rf"""(?ix)
            (?:hsl|hsla)?\s*(?:\(?\s*{hsl_part}
                (?:(?:\s*{fix_sep}\s*)((?:0*(?:0?\.[0-9]+|1\.0+|[0-9]+\.[0-9]+|[0-9]+))))?
            \s*\)?)"""
            if allow_alpha
            else rf"(?ix)(?:hsl|hsla)?\s*(?:\(?\s*{hsl_part}\s*\)?)"
        )

    @staticmethod
    def hexa_str(allow_alpha: bool = True) -> str:
        """Matches a HEXA color inside a string.\n
        ----------------------------------------------------------------------
        The HEXA color can be in the formats (prefix `#`, `0x` or no prefix):
        - `RGB`
        - `RGBA` (if `allow_alpha=True`)
        - `RRGGBB`
        - `RRGGBBAA` (if `allow_alpha=True`)\n
        #### Valid ranges:
        every channel from 0-9 and A-F (case insensitive)"""
        return (
            r"(?i)^(?:#|0x)?[0-9A-F]{8}|[0-9A-F]{6}|[0-9A-F]{4}|[0-9A-F]{3}$"
            if allow_alpha
            else r"(?i)^(?:#|0x)?[0-9A-F]{6}|[0-9A-F]{3}$"
        )



from . import __version__
from ._consts_ import DEFAULT
from .xx_format_codes import FormatCodes
from .xx_console import Console


def help_command():
    """Show some info about the library, with a brief explanation of how to use it."""
    color = {
        "lib": DEFAULT.color["ice"],
        "import": DEFAULT.color["red"],
        "class": DEFAULT.color["lavender"],
        "types": DEFAULT.color["lightblue"],
        "punctuators": DEFAULT.color["darkgray"],
    }
    FormatCodes.print(
        rf"""  [_|b|#7075FF]               __  __
  [b|#7075FF]  _  __ __  __/ / / /_  __  ___  __
  [b|#7075FF] | |/ // / / / / / __ \/ / / | |/ /
  [b|#7075FF] > , </ /_/ / /_/ /_/ / /_/ /> , <
  [b|#7075FF]/_/|_|\____/\__/\____/\____//_/|_|  [*|BG:{DEFAULT.color['gray']}|#000] v[b]{__version__} [*]

  [i|{DEFAULT.color['coral']}]A TON OF COOL FUNCTIONS, YOU NEED![*]

  [b|#75A2FF]Usage:[*]
    [{color['punctuators']}]# GENERAL LIBRARY[*]
    [{color['import']}]import [{color['lib']}]xulbux [{color['import']}]as [{color['lib']}]xx[*]
    [{color['punctuators']}]# CUSTOM TYPES[*]
    [{color['import']}]from [{color['lib']}]xulbux [{color['import']}]import [{color['lib']}]rgba[{color['punctuators']}], [{color['lib']}]hsla[{color['punctuators']}], [{color['lib']}]hexa[*]

  [b|#75A2FF]Includes:[*]
    [dim](•) CUSTOM TYPES:
       [dim](•) [{color['class']}]rgba[{color['punctuators']}]/([i|{color['types']}]int[_|{color['punctuators']}],[i|{color['types']}]int[_|{color['punctuators']}],[i|{color['types']}]int[_|{color['punctuators']}],[i|{color['types']}]float[_|{color['punctuators']}])[*]
       [dim](•) [{color['class']}]hsla[{color['punctuators']}]/([i|{color['types']}]int[_|{color['punctuators']}],[i|{color['types']}]int[_|{color['punctuators']}],[i|{color['types']}]int[_|{color['punctuators']}],[i|{color['types']}]float[_|{color['punctuators']}])[*]
       [dim](•) [{color['class']}]hexa[{color['punctuators']}]/([i|{color['types']}]str[_|{color['punctuators']}]|[i|{color['types']}]int[_|{color['punctuators']}])[*]
    [dim](•) PATH OPERATIONS          [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]Path[*]
    [dim](•) FILE OPERATIONS          [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]File[*]
    [dim](•) JSON FILE OPERATIONS     [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]Json[*]
    [dim](•) SYSTEM ACTIONS           [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]System[*]
    [dim](•) MANAGE THE ENV PATH VAR  [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]EnvPath[*]
    [dim](•) CONSOLE LOG AND ACTIONS  [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]Console[*]
    [dim](•) EASY PRETTY PRINTING     [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]FormatCodes[*]
    [dim](•) WORKING WITH COLORS      [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]Color[*]
    [dim](•) DATA OPERATIONS          [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]Data[*]
    [dim](•) STRING OPERATIONS        [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]String[*]
    [dim](•) CODE STRING OPERATIONS   [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]Code[*]
    [dim](•) REGEX PATTERN TEMPLATES  [{color['lib']}]xx[{color['punctuators']}].[{color['class']}]Regex[*]
  [_]
  [dim](Press any key to exit...)
  """,
        default_color=DEFAULT.text_color,
    )
    Console.pause_exit(pause=True)



from .xx_data import Data
from .xx_file import File

import json as _json
import os as _os


class Json:

    @staticmethod
    def read(
        json_file: str,
        comment_start: str = ">>",
        comment_end: str = "<<",
        return_original: bool = False,
    ) -> dict | tuple[dict, dict]:
        """Read JSON files, ignoring comments.\n
        -------------------------------------------------------------------------
        If only `comment_start` is found at the beginning of an item,
        the whole item is counted as a comment and therefore ignored.
        If `comment_start` and `comment_end` are found inside an item,
        the the section from `comment_start` to `comment_end` is ignored.
        If `return_original` is set to `True`, the original JSON is returned
        additionally. (returns: `[processed_json, original_json]`)"""
        if not json_file.endswith(".json"):
            json_file += ".json"
        file_path = File.extend_or_make_path(json_file, prefer_base_dir=True)
        with open(file_path, "r") as f:
            content = f.read()
        try:
            data = _json.loads(content)
        except _json.JSONDecodeError as e:
            raise ValueError(f"Error parsing JSON in '{file_path}':  {str(e)}")
        processed_data = Data.remove_comments(data, comment_start, comment_end)
        if not processed_data:
            raise ValueError(f"The JSON file '{file_path}' is empty or contains only comments.")
        return (processed_data, data) if return_original else processed_data

    @staticmethod
    def create(
        content: dict,
        new_file: str = "config",
        indent: int = 2,
        compactness: int = 1,
        force: bool = False,
    ) -> str:
        if not new_file.endswith(".json"):
            new_file += ".json"
        file_path = File.extend_or_make_path(new_file, prefer_base_dir=True)
        if _os.path.exists(file_path) and not force:
            with open(file_path, "r", encoding="utf-8") as existing_f:
                existing_content = _json.load(existing_f)
                if existing_content == content:
                    raise FileExistsError("Already created this file. (nothing changed)")
            raise FileExistsError("File already exists.")
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(Data.to_str(content, indent, compactness, as_json=True))
        full_path = _os.path.abspath(file_path)
        return full_path

    @staticmethod
    def update(
        json_file: str,
        update_values: str | list[str],
        comment_start: str = ">>",
        comment_end: str = "<<",
        sep: tuple[str, str] = ("->", "::"),
    ) -> None:
        """Function to easily update single/multiple values inside JSON files.\n
        ------------------------------------------------------------------------------------------------------
        The param `json_file` is the path to the JSON file or just the name of the JSON file to be updated.\n
        ------------------------------------------------------------------------------------------------------
        The param `update_values` is a sort of path (or a list of paths) to the value/s to be updated, with
        the new value at the end of the path.\n
        In this example:
        ```python
        {
          'healthy': {
            'fruit': ['apples', 'bananas', 'oranges'],
            'vegetables': ['carrots', 'broccoli', 'celery']
          }
        }
        ```
        ... if you want to change the value of `'apples'` to `'strawberries'`, `update_values` would be
        `healthy->fruit->apples::strawberries` or if you don't know that the value to update is `apples` you
        can also use the position of the value, so `healthy->fruit->0::strawberries`.\n
        ⇾ If the path from `update_values` doesn't exist, it will be created.\n
        ------------------------------------------------------------------------------------------------------
        If only `comment_start` is found at the beginning of an item, the whole item is counted as a comment
        and therefore ignored. If `comment_start` and `comment_end` are found inside an item, the the section
        from `comment_start` to `comment_end` is ignored."""
        if isinstance(update_values, str):
            update_values = [update_values]
        valid_entries = [
            (parts[0].strip(), parts[1])
            for update_value in update_values
            if len(parts := update_value.split(str(sep[1]).strip())) == 2
        ]
        value_paths, new_values = zip(*valid_entries) if valid_entries else ([], [])
        processed_data, data = Json.read(json_file, comment_start, comment_end, return_original=True)
        update = []
        for value_path, new_value in zip(value_paths, new_values):
            path_id = Data.get_path_id(processed_data, value_path)
            update.append(f"{path_id}::{new_value}")
        updated = Data.set_value_by_path_id(data, update)
        Json.create(updated, json_file, force=True)



from xulbux import Data


# ! DON'T CHANGE THIS DATA !
d_comments = {
    "key1": [
        ">> COMMENT IN THE BEGINNING OF THE STRING <<  value1",
        "value2  >> COMMENT IN THE END OF THE STRING",
        "val>> COMMENT IN THE MIDDLE OF THE STRING <<ue3",
        ">> FULL VALUE IS A COMMENT  value4",
    ],
    ">> FULL KEY + ALL ITS VALUES ARE A COMMENT  key2": ["value", "value", "value"],
    "key3": ">> ALL THE KEYS VALUES ARE COMMENTS  value",
}

d1_equal = {
    "key1": ["value1", "value2", "value3", ["value1", "value2", "value3"]],
    "key2": ["value1", "value2", "value3", ["value1", "value2", "value3"]],
    "key3": "value",
}
d2_equal = {
    "key1": ["value1", "value2", "value3", ["value1", "value2", "value3"]],
    "key2": ["value1", "value2", "value3", ["value1", "value2", "value3"]],
    "key3": "CHANGED value",
}

d1_path_id = {"healthy": {"fruit": ["apples", "bananas", "oranges"], "vegetables": ["carrots", "broccoli", "celery"]}}
d2_path_id = {"school": {"material": ["pencil", "paper", "rubber"], "subjects": ["math", "science", "history"]}}


def test_remove_comments():
    assert Data.remove_comments(d_comments, comment_sep="__") == {
        "key1": ["value1", "value2", "val__ue3"],
        "key3": None,
    }


def test_is_equal():
    assert Data.is_equal(d1_equal, d2_equal) == False
    assert Data.is_equal(d1_equal, d2_equal, ignore_paths="key3") == True


def test_path_id():
    id1, id2 = Data.get_path_id(d1_path_id, ["healthy->fruit->bananas", "healthy->vegetables->2"])
    assert id1 == "1>001"
    assert id2 == "1>012"
    assert Data.get_value_by_path_id(d1_path_id, id1) == "bananas"
    assert Data.get_value_by_path_id(d1_path_id, id2) == "celery"
    assert Data.set_value_by_path_id(d2_path_id, ["1>001::NEW1", "1>012::NEW2"]) == {
        "school": {"material": ["pencil", "NEW1", "rubber"], "subjects": ["math", "science", "NEW2"]}
    }



class DEFAULT:

    text_color: str = "#95B5FF"
    color: dict[str, str] = {
        "white": "#F1F2FF",
        "lightgray": "#B6B7C0",
        "gray": "#7B7C8D",
        "darkgray": "#67686C",
        "black": "#202125",
        "red": "#FF606A",
        "coral": "#FF7069",
        "orange": "#FF876A",
        "tangerine": "#FF9962",
        "gold": "#FFAF60",
        "yellow": "#FFD260",
        "green": "#7EE787",
        "teal": "#50EAAF",
        "cyan": "#3EE6DE",
        "ice": "#77EFEF",
        "lightblue": "#60AAFF",
        "blue": "#8085FF",
        "lavender": "#9B7DFF",
        "purple": "#AD68FF",
        "magenta": "#C860FF",
        "pink": "#EE60BB",
        "rose": "#FF6090",
    }


class CHARS:

    # CODE TO SIGNAL, ALL CHARACTERS ARE ALLOWED
    all = "<*allowed>"

    # DIGIT SETS
    digits = "0123456789"
    float_digits = digits + "."
    hex_digits = digits + "#abcdefABCDEF"

    # LETTER CATEGORIES
    lowercase = "abcdefghijklmnopqrstuvwxyz"
    lowercase_extended = lowercase + "äëïöüÿàèìòùáéíóúýâêîôûãñõåæç"
    uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    uppercase_extended = uppercase + "ÄËÏÖÜÀÈÌÒÙÁÉÍÓÚÝÂÊÎÔÛÃÑÕÅÆÇß"

    # COMBINED LETTER SETS
    letters = lowercase + uppercase
    letters_extended = lowercase_extended + uppercase_extended

    # ASCII sets
    special_ascii = " !\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"
    special_ascii_extended = special_ascii + "ø£Ø×ƒªº¿®¬½¼¡«»░▒▓│┤©╣║╗╝¢¥┐└┴┬├─┼╚╔╩╦╠═╬¤ðÐı┘┌█▄¦▀µþÞ¯´≡­±‗¾¶§÷¸°¨·¹³²■ "
    standard_ascii = special_ascii + digits + letters
    full_ascii = special_ascii_extended + digits + letters_extended


class ANSI:

    global CHAR, START, SEP, END

    escaped_char = "\\x1b"
    CHAR = char = "\x1b"
    START = start = "["
    SEP = sep = ";"
    END = end = "m"
    default_color_modifiers = {"lighten": "+l", "darken": "-d"}

    def seq(parts: int = 1) -> str:
        """Generate an ANSI sequence with `parts` amount of placeholders."""
        return CHAR + START + SEP.join(["{}" for _ in range(parts)]) + END

    seq_color: str = CHAR + START + "38" + SEP + "2" + SEP + "{}" + SEP + "{}" + SEP + "{}" + END
    seq_bg_color: str = CHAR + START + "48" + SEP + "2" + SEP + "{}" + SEP + "{}" + SEP + "{}" + END

    color_map: list[str] = [
        ########### DEFAULT CONSOLE COLOR NAMES ############
        "black",
        "red",
        "green",
        "yellow",
        "blue",
        "magenta",
        "cyan",
        "white",
    ]

    codes_map: dict[str | tuple[str], int] = {
        ################# SPECIFIC RESETS ##################
        "_": 0,
        ("_bold", "_b"): 22,
        ("_dim", "_d"): 22,
        ("_italic", "_i"): 23,
        ("_underline", "_u"): 24,
        ("_double-underline", "_du"): 24,
        ("_inverse", "_invert", "_in"): 27,
        ("_hidden", "_hide", "_h"): 28,
        ("_strikethrough", "_s"): 29,
        ("_color", "_c"): 39,
        ("_background", "_bg"): 49,
        ################### TEXT STYLES ####################
        ("bold", "b"): 1,
        ("dim", "d"): 2,
        ("italic", "i"): 3,
        ("underline", "u"): 4,
        ("inverse", "invert", "in"): 7,
        ("hidden", "hide", "h"): 8,
        ("strikethrough", "s"): 9,
        ("double-underline", "du"): 21,
        ################## DEFAULT COLORS ##################
        "black": 30,
        "red": 31,
        "green": 32,
        "yellow": 33,
        "blue": 34,
        "magenta": 35,
        "cyan": 36,
        "white": 37,
        ############## BRIGHT DEFAULT COLORS ###############
        "br:black": 90,
        "br:red": 91,
        "br:green": 92,
        "br:yellow": 93,
        "br:blue": 94,
        "br:magenta": 95,
        "br:cyan": 96,
        "br:white": 97,
        ############ DEFAULT BACKGROUND COLORS #############
        "bg:black": 40,
        "bg:red": 41,
        "bg:green": 42,
        "bg:yellow": 43,
        "bg:blue": 44,
        "bg:magenta": 45,
        "bg:cyan": 46,
        "bg:white": 47,
        ######### BRIGHT DEFAULT BACKGROUND COLORS #########
        "bg:br:black": 100,
        "bg:br:red": 101,
        "bg:br:green": 102,
        "bg:br:yellow": 103,
        "bg:br:blue": 104,
        "bg:br:magenta": 105,
        "bg:br:cyan": 106,
        "bg:br:white": 107,
    }



"""
Methods to transform formatting codes to ANSI and use them for pretty console output:
- `FormatCodes.print()` (print a special format-codes containing string)
- `FormatCodes.input()` (input with a special format-codes containing prompt)
- `FormatCodes.to_ansi()` (transform all special format-codes into ANSI codes in a string)\n
------------------------------------------------------------------------------------------------------------------------------------
### The Easy Formatting

First, let's take a look at a small example of what a highly styled print string with formatting could look like using this module:
```regex
This here is just unformatted text. [b|u|br:blue](Next we have text that is bright blue + bold + underlined.)\\n
[#000|bg:#F67](Then there's also black text with a red background.) And finally the ([i](boring)) plain text again.
```

How all of this exactly works is explained in the sections below. 🠫

------------------------------------------------------------------------------------------------------------------------------------
#### Formatting Codes and Keys

In this module, you can apply styles and colors using simple formatting codes.
These formatting codes consist of one or multiple different formatting keys in between square brackets.

If a formatting code is placed in a print-string, the formatting of that code will be applied to everything behind it until its
formatting is reset. If applying multiple styles and colors in the same place, instead of writing the formatting keys all into
separate brackets (e.g. `[x][y][z]`), they can also be put in a single pair of brackets, separated by pipes (e.g. `[x|y|z]`).

A list of all possible formatting keys can be found under all possible formatting keys.

------------------------------------------------------------------------------------------------------------------------------------
#### Auto Resetting Formatting Codes

Certain formatting can automatically be reset, behind a certain amount of text, just like shown in the following example:
```regex
This is plain text, [br:blue](which is bright blue now.) Now it was automatically reset to plain again.
```

This will only reset formatting codes, that have a specific reset listed below.
That means if you use it where another formatting is already applied, that formatting is still there after the automatic reset:
```regex
[cyan]This is cyan text, [dim](which is dimmed now.) Now it's not dimmed any more but still cyan.
```

If you want to ignore the auto-reset functionality of `()` brackets, you can put a `\\` or `/` between them and
the formatting code:
```regex
[cyan]This is cyan text, [u]/(which is underlined now.) And now it is still underlined and cyan.
```

------------------------------------------------------------------------------------------------------------------------------------
#### All possible Formatting Keys

- RGB colors:
  Change the text color directly with an RGB color inside the square brackets. (With or without `rgb()` brackets doesn't matter.)
  Examples:
  - `[rgb(115, 117, 255)]`
  - `[(255, 0, 136)]`
  - `[255, 0, 136]`
- HEX colors:
  Change the text color directly with a HEX color inside the square brackets. (If `RGB` or `RRGGBB` formatting code is used,
  and if there's a `#` or `0x` prefix, doesn't matter.)
  Examples:
  - `[#7788FF]`
  - `[7788FF]`
  - `[#78F]`
  - `[78F]`
- background RGB / HEX colors:
  Change the background color directly with an RGB or HEX color inside the square brackets, using the `background:` `BG:` prefix.
  (Same RGB / HEX formatting code rules as for text color.)
  Examples:
  - `[background:rgb(115, 117, 255)]`
  - `[BG:(255, 0, 136)]`
  - `[background:#7788FF]`
  - `[BG:#78F]`
- standard console colors:
  Change the text color to one of the standard console colors by just writing the color name in the square brackets.
  - `[black]`
  - `[red]`
  - `[green]`
  - `[yellow]`
  - `[blue]`
  - `[magenta]`
  - `[cyan]`
  - `[white]`
- bright console colors:
  Use the prefix `bright:` `BR:` to use the bright variant of the standard console color.
  Examples:
  - `[bright:black]` `[BR:black]`
  - `[bright:red]` `[BR:red]`
  - ...
- Background console colors:
  Use the prefix `background:` `BG:` to set the background to a standard console color. (Not all consoles support bright
  standard colors.)
  Examples:
  - `[background:black]` `[BG:black]`
  - `[background:red]` `[BG:red]`
  - ...
- Bright background console colors:
  Combine the prefixes `background:` / `BG:` and `bright:` / `BR:` to set the background to a bright console color.
  (The order of the prefixes doesn't matter.)
  Examples:
  - `[background:bright:black]` `[BG:BR:black]`
  - `[background:bright:red]` `[BG:BR:red]`
  - ...
- Text styles:
  Use the built-in text formatting to change the style of the text. There are long and short forms for each formatting code.
  (Not all consoles support all text styles.)
  - `[bold]` `[b]`
  - `[dim]`
  - `[italic]` `[i]`
  - `[underline]` `[u]`
  - `[inverse]` `[invert]` `[in]`
  - `[hidden]` `[hide]` `[h]`
  - `[strikethrough]` `[s]`
  - `[double-underline]` `[du]`
- Specific reset:
  Use these reset codes to remove a specific style, color or background. Again, there are long and
  short forms for each reset code.
  - `[_bold]` `[_b]`
  - `[_dim]`
  - `[_italic]` `[_i]`
  - `[_underline]` `[_u]`
  - `[_inverse]` `[_invert]` `[_in]`
  - `[_hidden]` `[_hide]` `[_h]`
  - `[_strikethrough]` `[_s]`
  - `[_double-underline]` `[_du]`
  - `[_color]` `[_c]`
  - `[_background]` `[_bg]`
- Total reset:
  This will reset all previously applied formatting codes.
  - `[_]`

------------------------------------------------------------------------------------------------------------------------------------
#### Additional Formatting Codes when a `default_color` is set

1. `[*]` resets everything, just like `[_]`, but the text color will remain in `default_color`
2. `[*color]` will reset the text color, just like `[_color]`, but then also make it `default_color`
3. `[default]` will just color the text in `default_color`
4. `[background:default]` `[BG:default]` will color the background in `default_color`

Unlike the standard console colors, the default color can be changed by using the following modifiers:

- `[l]` will lighten the `default_color` text by `brightness_steps`%
- `[ll]` will lighten the `default_color` text by `2 × brightness_steps`%
- `[lll]` will lighten the `default_color` text by `3 × brightness_steps`%
- ... etc. Same thing for darkening:
- `[d]` will darken the `default_color` text by `brightness_steps`%
- `[dd]` will darken the `default_color` text by `2 × brightness_steps`%
- `[ddd]` will darken the `default_color` text by `3 × brightness_steps`%
- ... etc.
Per default, you can also use `+` and `-` to get lighter and darker `default_color` versions.
"""

from ._consts_ import ANSI
from .xx_string import String
from .xx_regex import Regex
from .xx_color import *

import ctypes as _ctypes
import regex as _rx
import sys as _sys
import re as _re

_CONSOLE_ANSI_CONFIGURED = False

_PREFIX = {
    "BG": {"background", "bg"},
    "BR": {"bright", "br"},
}
_PREFIX_RX = {
    "BG": rf"(?:{'|'.join(_PREFIX['BG'])})\s*:",
    "BR": rf"(?:{'|'.join(_PREFIX['BR'])})\s*:",
}
_COMPILED = {  # PRECOMPILE REGULAR EXPRESSIONS
    "*": _re.compile(r"\[\s*([^]_]*?)\s*\*\s*([^]_]*?)\]"),
    "*color": _re.compile(r"\[\s*([^]_]*?)\s*\*color\s*([^]_]*?)\]"),
    "ansi_seq": _re.compile(ANSI.char + r"(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])"),
    "formatting": _rx.compile(
        Regex.brackets("[", "]", is_group=True)
        + r"(?:\s*([/\\]?)\s*"
        + Regex.brackets("(", ")", is_group=True, ignore_in_strings=False)
        + r")?"
    ),
    "bg?_default": _re.compile(r"(?i)((?:" + _PREFIX_RX["BG"] + r")?)\s*default"),
    "bg_default": _re.compile(r"(?i)" + _PREFIX_RX["BG"] + r"\s*default"),
    "modifier": _re.compile(
        r"(?i)((?:BG\s*:)?)\s*("
        + "|".join(
            [f"{_re.escape(m)}+" for m in ANSI.default_color_modifiers["lighten"] + ANSI.default_color_modifiers["darken"]]
        )
        + r")$"
    ),
    "rgb": _re.compile(
        r"(?i)^\s*(" + _PREFIX_RX["BG"] + r")?\s*(?:rgb|rgba)?\s*\(?\s*(\d{1,3})\s*,\s*(\d{1,3})\s*,\s*(\d{1,3})\s*\)?\s*$"
    ),
    "hex": _re.compile(r"(?i)^\s*(" + _PREFIX_RX["BG"] + r")?\s*(?:#|0x)?([0-9A-F]{6}|[0-9A-F]{3})\s*$"),
}


class FormatCodes:

    @staticmethod
    def print(
        *values: object,
        default_color: rgba | hexa = None,
        brightness_steps: int = 20,
        sep: str = " ",
        end: str = "\n",
        flush: bool = True,
    ) -> None:
        """A print function, whose print values can be formatted using formatting codes.\n
        -----------------------------------------------------------------------------------
        For exact information about how to use special formatting codes, see the
        `xx_format_codes` module documentation."""
        FormatCodes.__config_console()
        _sys.stdout.write(FormatCodes.to_ansi(sep.join(map(str, values)) + end, default_color, brightness_steps))
        if flush:
            _sys.stdout.flush()

    @staticmethod
    def input(
        prompt: object = "",
        default_color: rgba | hexa = None,
        brightness_steps: int = 20,
        reset_ansi: bool = False,
    ) -> str:
        """An input, whose prompt can be formatted using formatting codes.\n
        -------------------------------------------------------------------------
        If `reset_ansi` is true, all ANSI formatting will be reset, after the
        user confirms the input and the program continues to run.\n
        -------------------------------------------------------------------------
        For exact information about how to use special formatting codes, see the
        `xx_format_codes` module documentation."""
        FormatCodes.__config_console()
        user_input = input(FormatCodes.to_ansi(str(prompt), default_color, brightness_steps))
        if reset_ansi:
            _sys.stdout.write(f"{ANSI.char}[0m")
        return user_input

    @staticmethod
    def to_ansi(
        string: str,
        default_color: rgba | hexa = None,
        brightness_steps: int = 20,
        _default_start: bool = True,
    ) -> str:
        """Convert the formatting codes inside a string to ANSI formatting.\n
        -------------------------------------------------------------------------
        For exact information about how to use special formatting codes, see the
        `xx_format_codes` module documentation."""
        if not isinstance(string, str):
            string = str(string)
        if Color.is_valid_rgba(default_color, False):
            use_default = True
        elif Color.is_valid_hexa(default_color, False):
            use_default, default_color = True, Color.to_rgba(default_color)
        else:
            use_default = False
        if use_default:
            string = _COMPILED["*"].sub(r"[\1_|default\2]", string)  # REPLACE `[…|*|…]` WITH `[…|_|default|…]`
            string = _COMPILED["*color"].sub(r"[\1default\2]", string)  # REPLACE `[…|*color|…]` WITH `[…|default|…]`

        def is_valid_color(color: str) -> bool:
            return color in ANSI.color_map or Color.is_valid_rgba(color) or Color.is_valid_hexa(color)

        def replace_keys(match: _re.Match) -> str:
            formats = match.group(1)
            escaped = match.group(2)
            auto_reset_txt = match.group(3)
            if auto_reset_txt and auto_reset_txt.count("[") > 0 and auto_reset_txt.count("]") > 0:
                auto_reset_txt = FormatCodes.to_ansi(auto_reset_txt, default_color, brightness_steps, False)
            if not formats:
                return match.group(0)
            if formats.count("[") > 0 and formats.count("]") > 0:
                formats = FormatCodes.to_ansi(formats, default_color, brightness_steps, False)
            format_keys = [k.strip() for k in formats.split("|") if k.strip()]
            ansi_formats = [
                r if (r := FormatCodes.__get_replacement(k, default_color, brightness_steps)) != k else f"[{k}]"
                for k in format_keys
            ]
            if auto_reset_txt and not escaped:
                reset_keys = []
                for k in format_keys:
                    k_lower = k.lower()
                    k_set = set(k_lower.split(":"))
                    if _PREFIX["BG"] & k_set and len(k_set) <= 3:
                        if k_set & _PREFIX["BR"]:
                            for i in range(len(k)):
                                if is_valid_color(k[i:]):
                                    reset_keys.extend(["_bg", "_color"])
                                    break
                        else:
                            for i in range(len(k)):
                                if is_valid_color(k[i:]):
                                    reset_keys.append("_bg")
                                    break
                    elif is_valid_color(k) or any(
                        k_lower.startswith(pref_colon := f"{prefix}:") and is_valid_color(k[len(pref_colon) :])
                        for prefix in _PREFIX["BR"]
                    ):
                        reset_keys.append("_color")
                    else:
                        reset_keys.append(f"_{k}")
                ansi_resets = [
                    r
                    for k in reset_keys
                    if (r := FormatCodes.__get_replacement(k, default_color, brightness_steps)).startswith(
                        f"{ANSI.char}{ANSI.start}"
                    )
                ]
            else:
                ansi_resets = []
            if not (len(ansi_formats) == 1 and ansi_formats[0].count(f"{ANSI.char}{ANSI.start}") >= 1) and not all(
                f.startswith(f"{ANSI.char}{ANSI.start}") for f in ansi_formats
            ):
                return match.group(0)
            return (
                "".join(ansi_formats)
                + (
                    f"({FormatCodes.to_ansi(auto_reset_txt, default_color, brightness_steps, False)})"
                    if escaped and auto_reset_txt
                    else auto_reset_txt if auto_reset_txt else ""
                )
                + ("" if escaped else "".join(ansi_resets))
            )

        string = "\n".join(_COMPILED["formatting"].sub(replace_keys, line) for line in string.split("\n"))
        return (FormatCodes.__get_default_ansi(default_color) if _default_start else "") + string if use_default else string

    @staticmethod
    def escape_ansi(ansi_string: str) -> str:
        """Escapes all ANSI codes in the string, so they are visible when output to the console."""
        return ansi_string.replace(ANSI.char, ANSI.escaped_char)

    @staticmethod
    def remove_ansi(ansi_string: str) -> str:
        """Removes all ANSI codes from the string."""
        return _COMPILED["ansi_seq"].sub("", ansi_string)

    @staticmethod
    def remove_formatting(string: str) -> str:
        """Removes all formatting codes from the string."""
        return _COMPILED["ansi_seq"].sub("", FormatCodes.to_ansi(string))

    @staticmethod
    def __config_console() -> None:
        """Configure the console to be able to interpret ANSI formatting."""
        global _CONSOLE_ANSI_CONFIGURED
        if not _CONSOLE_ANSI_CONFIGURED:
            _sys.stdout.flush()
            kernel32 = _ctypes.windll.kernel32
            h = kernel32.GetStdHandle(-11)
            mode = _ctypes.c_ulong()
            kernel32.GetConsoleMode(h, _ctypes.byref(mode))
            kernel32.SetConsoleMode(h, mode.value | 0x0004)
            _CONSOLE_ANSI_CONFIGURED = True

    @staticmethod
    def __get_default_ansi(
        default_color: tuple,
        format_key: str = None,
        brightness_steps: int = None,
        _modifiers: tuple[str, str] = (ANSI.default_color_modifiers["lighten"], ANSI.default_color_modifiers["darken"]),
    ) -> str | None:
        """Get the `default_color` and lighter/darker versions of it as ANSI code."""
        if not brightness_steps or (format_key and _COMPILED["bg?_default"].search(format_key)):
            return (ANSI.seq_bg_color if format_key and _COMPILED["bg_default"].search(format_key) else ANSI.seq_color).format(
                *default_color[:3]
            )
        if not (format_key in _modifiers[0] or format_key in _modifiers[1]):
            return None
        match = _COMPILED["modifier"].match(format_key)
        if not match:
            return None
        is_bg, modifiers = match.groups()
        adjust = 0
        for mod in _modifiers[0] + _modifiers[1]:
            adjust = String.single_char_repeats(modifiers, mod)
            if adjust and adjust > 0:
                modifiers = mod
                break
        if adjust == 0:
            return None
        elif modifiers in _modifiers[0]:
            new_rgb = Color.adjust_lightness(default_color, (brightness_steps / 100) * adjust)
        elif modifiers in _modifiers[1]:
            new_rgb = Color.adjust_lightness(default_color, -(brightness_steps / 100) * adjust)
        return (ANSI.seq_bg_color if is_bg else ANSI.seq_color).format(*new_rgb[:3])

    @staticmethod
    def __get_replacement(format_key: str, default_color: rgba = None, brightness_steps: int = 20) -> str:
        """Gives you the corresponding ANSI code for the given format key.
        If `default_color` is not `None`, the text color will be `default_color` if all formats
        are reset or you can get lighter or darker version of `default_color` (also as BG)"""
        use_default = default_color and Color.is_valid_rgba(default_color, False)
        _format_key, format_key = format_key, FormatCodes.__normalize_key(format_key)  # NORMALIZE KEY AND SAVE ORIGINAL
        if use_default:
            if new_default_color := FormatCodes.__get_default_ansi(default_color, format_key, brightness_steps):
                return new_default_color
        for map_key in ANSI.codes_map:
            if (isinstance(map_key, tuple) and format_key in map_key) or format_key == map_key:
                return ANSI.seq().format(
                    next(
                        (
                            v
                            for k, v in ANSI.codes_map.items()
                            if format_key == k or (isinstance(k, tuple) and format_key in k)
                        ),
                        None,
                    )
                )
        rgb_match = _re.match(_COMPILED["rgb"], format_key)
        hex_match = _re.match(_COMPILED["hex"], format_key)
        try:
            if rgb_match:
                is_bg = rgb_match.group(1)
                r, g, b = map(int, rgb_match.groups()[1:])
                if Color.is_valid_rgba((r, g, b)):
                    return ANSI.seq_bg_color.format(r, g, b) if is_bg else ANSI.seq_color.format(r, g, b)
            elif hex_match:
                is_bg = hex_match.group(1)
                rgb = Color.to_rgba(hex_match.group(2))
                return (
                    ANSI.seq_bg_color.format(rgb[0], rgb[1], rgb[2])
                    if is_bg
                    else ANSI.seq_color.format(rgb[0], rgb[1], rgb[2])
                )
        except Exception:
            pass
        return _format_key

    @staticmethod
    def __normalize_key(format_key: str) -> str:
        """Normalizes the given format key."""
        k_parts = format_key.replace(" ", "").lower().split(":")
        prefix_str = "".join(
            f"{prefix_key.lower()}:"
            for prefix_key, prefix_values in _PREFIX.items()
            if any(k_part in prefix_values for k_part in k_parts)
        )
        return prefix_str + ":".join(
            part for part in k_parts if part not in {val for values in _PREFIX.values() for val in values}
        )



from xulbux._consts_ import ANSI
from xulbux import FormatCodes


black = ANSI.seq_color.format(0, 0, 0)
bg_red = f"{ANSI.char}{ANSI.start}{ANSI.codes_map['bg:red']}{ANSI.end}"
default = ANSI.seq_color.format(255, 255, 255)
orange = ANSI.seq_color.format(255, 136, 119)

bold = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('bold', 'b')]}{ANSI.end}"
invert = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('inverse', 'invert', 'in')]}{ANSI.end}"
italic = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('italic', 'i')]}{ANSI.end}"
underline = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('underline', 'u')]}{ANSI.end}"

reset = f"{ANSI.char}{ANSI.start}{ANSI.codes_map['_']}{ANSI.end}"
reset_bg = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_background', '_bg')]}{ANSI.end}"
reset_bold = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_bold', '_b')]}{ANSI.end}"
reset_color = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_color', '_c')]}{ANSI.end}"
reset_italic = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_italic', '_i')]}{ANSI.end}"
reset_invert = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_inverse', '_invert', '_in')]}{ANSI.end}"
reset_underline = f"{ANSI.char}{ANSI.start}{ANSI.codes_map[('_underline', '_u')]}{ANSI.end}"


def test_codes_to_ansi():
    assert (
        FormatCodes.to_ansi("[b|#000|bg:red](He[in](l)lo) [[i|u|#F87](world)][default]![_]", default_color="#FFF")
        == f"{default}{bold}{black}{bg_red}"
        + "He"
        + invert
        + "l"
        + reset_invert
        + "lo"
        + f"{reset_bold}{reset_color}{reset_bg}"
        + " ["
        + f"{italic}{underline}{orange}"
        + "world"
        + f"{reset_italic}{reset_underline}{reset_color}"
        + "]"
        + default
        + "!"
        + reset
    )



import subprocess as _subprocess
import platform as _platform
import ctypes as _ctypes
import time as _time
import sys as _sys
import os as _os


class ProcessNotFoundError(Exception):
    pass


class System:

    @staticmethod
    def is_elevated() -> bool:
        """Returns `True` if the current user is an admin and `False` otherwise."""
        try:
            if _os.name == "nt":
                return _ctypes.windll.shell32.IsUserAnAdmin() != 0
            elif _os.name == "posix":
                return _os.geteuid() == 0
        except:
            pass
        return False

    @staticmethod
    def restart(
        prompt: object = None,
        wait: int = 0,
        continue_program: bool = False,
        force: bool = False,
    ) -> None:
        """Starts a system restart:
        - `prompt` is the message to be displayed in the systems restart notification.
        - `wait` is the time to wait until restarting in seconds.
        - `continue_program` is whether to continue the current Python program after calling this function.
        - `force` is whether to force a restart even if other processes are still running."""
        system = _platform.system().lower()
        if system == "windows":
            if not force:
                output = _subprocess.check_output("tasklist", shell=True).decode()
                processes = [line.split()[0] for line in output.splitlines()[3:] if line.strip()]
                if len(processes) > 2:  # EXCLUDING THE PYTHON PROCESS AND CONSOLE
                    raise RuntimeError("Processes are still running. Use the parameter `force=True` to restart anyway.")
            if prompt:
                _os.system(f'shutdown /r /t {wait} /c "{prompt}"')
            else:
                _os.system("shutdown /r /t 0")
            if continue_program:
                print(f"Restarting in {wait} seconds...")
                _time.sleep(wait)
        elif system in ("linux", "darwin"):
            if not force:
                output = _subprocess.check_output(["ps", "-A"]).decode()
                processes = output.splitlines()[1:]  # EXCLUDE HEADER
                if len(processes) > 2:  # EXCLUDING THE PYTHON PROCESS AND PS
                    raise RuntimeError("Processes are still running. Use the parameter `force=True` to restart anyway.")
            if prompt:
                _subprocess.Popen(["notify-send", "System Restart", prompt])
                _time.sleep(wait)
            try:
                _subprocess.run(["sudo", "shutdown", "-r", "now"])
            except _subprocess.CalledProcessError:
                raise PermissionError("Failed to restart: insufficient privileges. Ensure sudo permissions are granted.")
            if continue_program:
                print(f"Restarting in {wait} seconds...")
                _time.sleep(wait)
        else:
            raise NotImplementedError(f"Restart not implemented for `{system}`")

    @staticmethod
    def check_libs(
        lib_names: list[str],
        install_missing: bool = False,
        confirm_install: bool = True,
    ) -> None | list[str]:
        """Checks if the given list of libraries are installed. If not:
        - If `install_missing` is `False` the missing libraries will be returned as a list.
        - If `install_missing` is `True` the missing libraries will be installed.
        - If `confirm_install` is `True` the user will first be asked if they want to install the missing libraries."""
        missing = []
        for lib in lib_names:
            try:
                __import__(lib)
            except ImportError:
                missing.append(lib)
        if not missing:
            return None
        elif not install_missing:
            return missing
        if confirm_install:
            print("The following required libraries are missing:")
            for lib in missing:
                print(f"- {lib}")
            if input("Do you want to install them now (Y/n):  ").strip().lower() not in ("", "y", "yes"):
                raise ImportError("Missing required libraries.")
        try:
            _subprocess.check_call([_sys.executable, "-m", "pip", "install"] + missing)
            return None
        except _subprocess.CalledProcessError:
            return missing

    @staticmethod
    def elevate(win_title: str | None = None, args: list | None = None) -> bool:
        """Attempts to start a new process with elevated privileges.\n
        ---------------------------------------------------------------------------------
        The param `win_title` is window the title of the elevated process.
        The param `args` is the arguments to be passed to the elevated process.\n
        After the elevated process started, the original process will exit.
        This means, that this method has to be run at the beginning of the program or
        will have to continue in a new window after elevation.\n
        ---------------------------------------------------------------------------------
        Returns `True` if the current process already has elevated privileges and raises
        a `PermissionError` if the user denied the elevation or the elevation failed."""
        if System.is_elevated():
            return True
        if _os.name == "nt":  # WINDOWS
            if win_title:
                args_str = f'-c "import ctypes; ctypes.windll.kernel32.SetConsoleTitleW(\\"{win_title}\\"); exec(open(\\"{_sys.argv[0]}\\").read())" {" ".join(args)}"'
            else:
                args_str = f'-c "exec(open(\\"{_sys.argv[0]}\\").read())" {" ".join(args)}'
            result = _ctypes.windll.shell32.ShellExecuteW(None, "runas", _sys.executable, args_str, None, 1)
            if result <= 32:
                raise PermissionError("Failed to launch elevated process.")
            else:
                _sys.exit(0)
        else:  # POSIX
            cmd = ["pkexec"]
            if win_title:
                cmd.extend(["--description", win_title])
            cmd.extend([_sys.executable] + _sys.argv[1:] + ([] if args is None else args))
            proc = _subprocess.Popen(cmd)
            proc.wait()
            if proc.returncode != 0:
                raise PermissionError("Process elevation was denied.")
            _sys.exit(0)



from .xx_string import String
from .xx_regex import Regex
from .xx_data import Data

import regex as _rx


class Code:

    @staticmethod
    def add_indent(code: str, indent: int) -> str:
        """Adds `indent` spaces at the beginning of each line."""
        indented_lines = [" " * indent + line for line in code.splitlines()]
        return "\n".join(indented_lines)

    @staticmethod
    def get_tab_spaces(code: str) -> int:
        """Will try to get the amount of spaces used for indentation."""
        code_lines = String.get_lines(code, remove_empty_lines=True)
        indents = [len(line) - len(line.lstrip()) for line in code_lines]
        non_zero_indents = [i for i in indents if i > 0]
        return min(non_zero_indents) if non_zero_indents else 0

    @staticmethod
    def change_tab_size(code: str, new_tab_size: int, remove_empty_lines: bool = False) -> str:
        """Replaces all tabs with `new_tab_size` spaces.\n
        ----------------------------------------------------------------------------------
        If `remove_empty_lines` is `True`, empty lines will be removed in the process."""
        code_lines = String.get_lines(code, remove_empty_lines=True)
        lines = code_lines if remove_empty_lines else String.get_lines(code)
        tab_spaces = Code.get_tab_spaces(code)
        if (tab_spaces == new_tab_size) or tab_spaces == 0:
            if remove_empty_lines:
                return "\n".join(code_lines)
            return code
        result = []
        for line in lines:
            stripped = line.lstrip()
            indent_level = (len(line) - len(stripped)) // tab_spaces
            new_indent = " " * (indent_level * new_tab_size)
            result.append(new_indent + stripped)
        return "\n".join(result)

    @staticmethod
    def get_func_calls(code: str) -> list:
        """Will try to get all function calls and return them as a list."""
        funcs = _rx.findall(r"(?i)" + Regex.func_call(), code)
        nested_func_calls = []
        for _, func_attrs in funcs:
            nested_calls = _rx.findall(r"(?i)" + Regex.func_call(), func_attrs)
            if nested_calls:
                nested_func_calls.extend(nested_calls)
        return Data.remove_duplicates(funcs + nested_func_calls)

    @staticmethod
    def is_js(code: str, funcs: list = ["__", "$t", "$lang"]) -> bool:
        """Will check if the code is very likely to be JavaScript."""
        funcs = "|".join(funcs)
        js_pattern = _rx.compile(
            Regex.outside_strings(
                r"""^(?:
            (\$[\w_]+)\s*                      # JQUERY-STYLE VARIABLES
            |(\$[\w_]+\s*\()                   # JQUERY-STYLE FUNCTION CALLS
            |(("""
                + funcs
                + r")"
                + Regex.brackets("()")
                + r"""\s*) # PREDEFINED FUNCTION CALLS
            |(\bfunction\s*\()                 # FUNCTION DECLARATIONS
            |(\b(var|let|const)\s+[\w_]+\s*=)  # VARIABLE DECLARATIONS
            |(\b(if|for|while|switch)\s*\()    # CONTROL STRUCTURES
            |(\b(return|throw)\s+)             # RETURN OR THROW STATEMENTS
            |(\bnew\s+[\w_]+\()                # OBJECT INSTANTIATION
            |(\b[\w_]+\s*=>\s*{)               # ARROW FUNCTIONS
            |(\b(true|false|null|undefined)\b) # JAVASCRIPT LITERALS
            |(\b(document|window|console)\.)   # BROWSER OBJECTS
            |(\b[\w_]+\.(forEach|map|filter|reduce)\() # ARRAY METHODS
            |(/[^/\n\r]*?/[gimsuy]*)           # REGULAR EXPRESSIONS
            |(===|!==|\+\+|--|\|\||&&)         # JAVASCRIPT-SPECIFIC OPERATORS
            |(\bclass\s+[\w_]+)                # CLASS DECLARATIONS
            |(\bimport\s+.*?from\s+)           # IMPORT STATEMENTS
            |(\bexport\s+(default\s+)?)        # EXPORT STATEMENTS
            |(\basync\s+function)              # ASYNC FUNCTIONS
            |(\bawait\s+)                      # AWAIT KEYWORD
            |(\btry\s*{)                       # TRY-CATCH BLOCKS
            |(\bcatch\s*\()
            |(\bfinally\s*{)
            |(\byield\s+)                      # GENERATOR FUNCTIONS
            |(\[.*?\]\s*=)                     # DESTRUCTURING ASSIGNMENT
            |(\.\.\.)                          # SPREAD OPERATOR
            |(==|!=|>=|<=|>|<)                 # COMPARISON OPERATORS
            |(\+=|-=|\*=|/=|%=|\*\*=)          # COMPOUND ASSIGNMENT OPERATORS
            |(\+|-|\*|/|%|\*\*)                # ARITHMETIC OPERATORS
            |(&|\||\^|~|<<|>>|>>>)             # BITWISE OPERATORS
            |(\?|:)                            # TERNARY OPERATOR
            |(\bin\b)                          # IN OPERATOR
            |(\binstanceof\b)                  # INSTANCEOF OPERATOR
            |(\bdelete\b)                      # DELETE OPERATOR
            |(\btypeof\b)                      # TYPEOF OPERATOR
            |(\bvoid\b)                        # VOID OPERATOR
        )[\s\S]*$"""
            ),
            _rx.VERBOSE | _rx.IGNORECASE,
        )
        return bool(js_pattern.fullmatch(code))



from .xx_string import String

from typing import TypeAlias, Union
import math as _math
import re as _re


DataStructure: TypeAlias = Union[list, tuple, set, frozenset, dict]


class Data:

    @staticmethod
    def chars_count(data: DataStructure) -> int:
        """The sum of all the characters amount including the keys in dictionaries."""
        if isinstance(data, dict):
            return sum(len(str(k)) + len(str(v)) for k, v in data.items())
        return sum(len(str(item)) for item in data)

    @staticmethod
    def strip(data: DataStructure) -> DataStructure:
        """Removes leading and trailing whitespaces from the data structure's items."""
        if isinstance(data, dict):
            return {k: Data.strip(v) for k, v in data.items()}
        return type(data)(map(Data.strip, data))

    @staticmethod
    def remove_empty_items(data: DataStructure, spaces_are_empty: bool = False) -> DataStructure:
        """Removes empty items from the data structure.
        If `spaces_are_empty` is true, it will count items with only spaces as empty."""
        if isinstance(data, dict):
            return {
                k: (
                    v
                    if not isinstance(v, (list, tuple, set, frozenset, dict))
                    else Data.remove_empty_items(v, spaces_are_empty)
                )
                for k, v in data.items()
                if not String.is_empty(v, spaces_are_empty)
            }
        if isinstance(data, (list, tuple, set, frozenset)):
            return type(data)(
                item
                for item in (
                    (
                        item
                        if not isinstance(item, (list, tuple, set, frozenset, dict))
                        else Data.remove_empty_items(item, spaces_are_empty)
                    )
                    for item in data
                    if not String.is_empty(item, spaces_are_empty)
                )
                if item not in ((), {}, set(), frozenset())
            )
        return data

    @staticmethod
    def remove_duplicates(data: DataStructure) -> DataStructure:
        """Removes all duplicates from the data structure."""
        if isinstance(data, dict):
            return {k: Data.remove_duplicates(v) for k, v in data.items()}
        if isinstance(data, (list, tuple)):
            return type(data)(
                Data.remove_duplicates(item) if isinstance(item, (list, tuple, set, frozenset, dict)) else item
                for item in dict.fromkeys(data)
            )
        if isinstance(data, (set, frozenset)):
            return type(data)(
                Data.remove_duplicates(item) if isinstance(item, (list, tuple, set, frozenset, dict)) else item
                for item in data
            )
        return data

    @staticmethod
    def remove_comments(
        data: DataStructure,
        comment_start: str = ">>",
        comment_end: str = "<<",
        comment_sep: str = "",
    ) -> DataStructure:
        """Remove comments from a list, tuple or dictionary.\n
        ----------------------------------------------------------------------------------------------------------------------
        - The `data` parameter is your list, tuple or dictionary, where the comments should get removed from.
        - The `comment_start` parameter is the string that marks the start of a comment inside `data`. (default: `>>`)
        - The `comment_end` parameter is the string that marks the end of a comment inside `data`. (default: `<<`)
        - The `comment_sep` parameter is a string with which a comment will be replaced, if it is in the middle of a value.\n
        ----------------------------------------------------------------------------------------------------------------------
        Examples:
        ```python
        data = {
            "key1": [
                ">> COMMENT IN THE BEGINNING OF THE STRING <<  value1",
                "value2  >> COMMENT IN THE END OF THE STRING",
                "val>> COMMENT IN THE MIDDLE OF THE STRING <<ue3",
                ">> FULL VALUE IS A COMMENT  value4"
            ],
            ">> FULL KEY + ALL ITS VALUES ARE A COMMENT  key2": [
                "value",
                "value",
                "value"
            ],
            "key3": ">> ALL THE KEYS VALUES ARE COMMENTS  value"
        }

        processed_data = Data.remove_comments(
            data,
            comment_start=">>",
            comment_end="<<",
            comment_sep="__"
        )
        ```\n
        ----------------------------------------------------------------------------------------------------------------------
        For this example, `processed_data` will be:
        ```python
        {
            "key1": [
                "value1",
                "value2",
                "val__ue3"
            ],
            "key3": None
        }
        ```\n
        - For `key1`, all the comments will just be removed, except at `value3` and `value4`:
          - `value3` The comment is removed and the parts left and right are joined through `comment_sep`.
          - `value4` The whole value is removed, since the whole value was a comment.
        - For `key2`, the key, including its whole values will be removed.
        - For `key3`, since all its values are just comments, the key will still exist, but with a value of `None`."""
        if comment_end:
            pattern = _re.compile(
                rf"^((?:(?!{_re.escape(comment_start)}).)*){_re.escape(comment_start)}(?:(?:(?!{_re.escape(comment_end)}).)*)(?:{_re.escape(comment_end)})?(.*?)$"
            )

        def process_string(s: str) -> str | None:
            if comment_end:
                match = pattern.match(s)
                if match:
                    start, end = match.group(1).strip(), match.group(2).strip()
                    return f"{start}{comment_sep if start and end else ''}{end}" or None
                return s.strip() or None
            else:
                return None if s.lstrip().startswith(comment_start) else s.strip() or None

        def process_item(item: any) -> any:
            if isinstance(item, dict):
                return {
                    k: v for k, v in ((process_item(key), process_item(value)) for key, value in item.items()) if k is not None
                }
            if isinstance(item, (list, tuple, set, frozenset)):
                processed = (v for v in map(process_item, item) if v is not None)
                return type(item)(processed)
            if isinstance(item, str):
                return process_string(item)
            return item

        return process_item(data)

    @staticmethod
    def is_equal(
        data1: DataStructure,
        data2: DataStructure,
        ignore_paths: str | list[str] = "",
        path_sep: str = "->",
        comment_start: str = ">>",
        comment_end: str = "<<",
    ) -> bool:
        """Compares two structures and returns `True` if they are equal and `False` otherwise.\n
        ⇾ Will not detect, if a key-name has changed, only if removed or added.\n
        --------------------------------------------------------------------------------------------
        Ignores the specified (found) key/s or item/s from `ignore_paths`. Comments are not ignored
        when comparing. `comment_start` and `comment_end` are only used to correctly recognize the
        keys in the `ignore_paths`.\n
        --------------------------------------------------------------------------------------------
        The paths from `ignore_paths` and the `path_sep` parameter work exactly the same way as for
        the function `Data.get_path_id()`. See its documentation for more details."""

        def process_ignore_paths(
            ignore_paths: str | list[str],
        ) -> list[list[str]]:
            if isinstance(ignore_paths, str):
                ignore_paths = [ignore_paths]
            return [path.split(path_sep) for path in ignore_paths if path]

        def compare(
            d1: DataStructure,
            d2: DataStructure,
            ignore_paths: list[list[str]],
            current_path: list[str] = [],
        ) -> bool:
            if any(current_path == path[: len(current_path)] for path in ignore_paths):
                return True
            if type(d1) != type(d2):
                return False
            if isinstance(d1, dict):
                if set(d1.keys()) != set(d2.keys()):
                    return False
                return all(compare(d1[key], d2[key], ignore_paths, current_path + [key]) for key in d1)
            if isinstance(d1, (list, tuple)):
                if len(d1) != len(d2):
                    return False
                return all(
                    compare(item1, item2, ignore_paths, current_path + [str(i)])
                    for i, (item1, item2) in enumerate(zip(d1, d2))
                )
            if isinstance(d1, (set, frozenset)):
                return d1 == d2
            return d1 == d2

        processed_data1 = Data.remove_comments(data1, comment_start, comment_end)
        processed_data2 = Data.remove_comments(data2, comment_start, comment_end)
        processed_ignore_paths = process_ignore_paths(ignore_paths)
        return compare(processed_data1, processed_data2, processed_ignore_paths)

    @staticmethod
    def get_path_id(
        data: DataStructure,
        value_paths: str | list[str],
        path_sep: str = "->",
        comment_start: str = ">>",
        comment_end: str = "<<",
        ignore_not_found: bool = False,
    ) -> str | list[str]:
        """Generates a unique ID based on the path to a specific value within a nested data structure.\n
        -------------------------------------------------------------------------------------------------
        The `data` parameter is the list, tuple, or dictionary, which the id should be generated for.\n
        -------------------------------------------------------------------------------------------------
        The param `value_path` is a sort of path (or a list of paths) to the value/s to be updated.
        In this example:
        ```python
        {
            "healthy": {
                "fruit": ["apples", "bananas", "oranges"],
                "vegetables": ["carrots", "broccoli", "celery"]
            }
        }
        ```
        ... if you want to change the value of `"apples"` to `"strawberries"`, the value path would be
        `healthy->fruit->apples` or if you don't know that the value is `"apples"` you can also use the
        index of the value, so `healthy->fruit->0`.\n
        -------------------------------------------------------------------------------------------------
        The comments marked with `comment_start` and `comment_end` will be removed, before trying to get
        the path id.\n
        -------------------------------------------------------------------------------------------------
        The `path_sep` param is the separator between the keys/indexes in the path (default is `->` just
        like in the example above).\n
        -------------------------------------------------------------------------------------------------
        If `ignore_not_found` is `True`, the function will return `None` if the value is not found
        instead of raising an error."""

        def process_path(path: str, data_obj: list | tuple | set | frozenset | dict) -> str | None:
            keys = path.split(path_sep)
            path_ids = []
            max_id_length = 0
            for key in keys:
                if isinstance(data_obj, dict):
                    if key.isdigit():
                        if ignore_not_found:
                            return None
                        raise TypeError(f"Key '{key}' is invalid for a dict type.")
                    try:
                        idx = list(data_obj.keys()).index(key)
                        data_obj = data_obj[key]
                    except (ValueError, KeyError):
                        if ignore_not_found:
                            return None
                        raise KeyError(f"Key '{key}' not found in dict.")
                elif isinstance(data_obj, (list, tuple, set, frozenset)):
                    try:
                        idx = int(key)
                        data_obj = list(data_obj)[idx]  # CONVERT TO LIST FOR INDEXING
                    except ValueError:
                        try:
                            idx = list(data_obj).index(key)
                            data_obj = list(data_obj)[idx]
                        except ValueError:
                            if ignore_not_found:
                                return None
                            raise ValueError(f"Value '{key}' not found in '{type(data_obj).__name__}'")
                else:
                    break
                path_ids.append(str(idx))
                max_id_length = max(max_id_length, len(str(idx)))
            if not path_ids:
                return None
            return f"{max_id_length}>{''.join(id.zfill(max_id_length) for id in path_ids)}"

        data = Data.remove_comments(data, comment_start, comment_end)
        if isinstance(value_paths, str):
            return process_path(value_paths, data)
        results = [process_path(path, data) for path in value_paths]
        return results if len(results) > 1 else results[0] if results else None

    @staticmethod
    def get_value_by_path_id(data: DataStructure, path_id: str, get_key: bool = False) -> any:
        """Retrieves the value from `data` using the provided `path_id`.\n
        -------------------------------------------------------------------------------------------------
        Input your `data` along with a `path_id` that was created before using `Data.get_path_id()`.
        If `get_key` is true and the final item is in a dict, it returns the key instead of the value.\n
        -------------------------------------------------------------------------------------------------
        The function will return the value (or key) from the path ID location, as long as the structure
        of `data` hasn't changed since creating the path ID to that value."""

        def get_nested(data: list | tuple | set | frozenset | dict, path: list[int], get_key: bool) -> any:
            parent = None
            for i, idx in enumerate(path):
                if isinstance(data, dict):
                    keys = list(data.keys())
                    if i == len(path) - 1 and get_key:
                        return keys[idx]
                    parent = data
                    data = data[keys[idx]]
                elif isinstance(data, (list, tuple, set, frozenset)):
                    if i == len(path) - 1 and get_key:
                        if parent is None or not isinstance(parent, dict):
                            raise ValueError("Cannot get key from a non-dict parent")
                        return next(key for key, value in parent.items() if value is data)
                    parent = data
                    data = list(data)[idx]  # CONVERT TO LIST FOR INDEXING
                else:
                    raise TypeError(f"Unsupported type '{type(data)}' at path '{path[:i+1]}'")
            return data

        return get_nested(data, Data.__sep_path_id(path_id), get_key)

    @staticmethod
    def set_value_by_path_id(
        data: DataStructure,
        update_values: str | list[str],
        sep: str = "::",
    ) -> list | tuple | dict:
        """Updates the value/s from `update_values` in the `data`.\n
        --------------------------------------------------------------------------------
        Input a list, tuple or dict as `data`, along with `update_values`, which is a
        path ID that was created before using `Data.get_path_id()`, together with the
        new value to be inserted where the path ID points to. The path ID and the new
        value are separated by `sep`, which per default is `::`.\n
        --------------------------------------------------------------------------------
        The value from path ID will be changed to the new value, as long as the
        structure of `data` hasn't changed since creating the path ID to that value."""

        def update_nested(
            data: list | tuple | set | frozenset | dict, path: list[int], value: any
        ) -> list | tuple | set | frozenset | dict:
            if len(path) == 1:
                if isinstance(data, dict):
                    keys = list(data.keys())
                    data = dict(data)
                    data[keys[path[0]]] = value
                elif isinstance(data, (list, tuple, set, frozenset)):
                    data = list(data)
                    data[path[0]] = value
                    data = type(data)(data)
            else:
                if isinstance(data, dict):
                    keys = list(data.keys())
                    key = keys[path[0]]
                    data = dict(data)
                    data[key] = update_nested(data[key], path[1:], value)
                elif isinstance(data, (list, tuple, set, frozenset)):
                    data = list(data)
                    data[path[0]] = update_nested(data[path[0]], path[1:], value)
                    data = type(data)(data)
            return data

        if isinstance(update_values, str):
            update_values = [update_values]
        valid_entries = [
            (parts[0].strip(), parts[1])
            for update_value in update_values
            if len(parts := update_value.split(str(sep).strip())) == 2
        ]
        if not valid_entries:
            raise ValueError(f"No valid update_values found: {update_values}")
        for path_id, new_val in valid_entries:
            path = Data.__sep_path_id(path_id)
            data = update_nested(data, path, new_val)
        return data

    @staticmethod
    def print(
        data: DataStructure,
        indent: int = 4,
        compactness: int = 1,
        max_width: int = 127,
        sep: str = ", ",
        end: str = "\n",
        as_json: bool = False,
    ) -> None:
        """Print nicely formatted data structures.\n
        ------------------------------------------------------------------------------
        The indentation spaces-amount can be set with with `indent`.
        There are three different levels of `compactness`:
        - `0` expands everything possible
        - `1` only expands if there's other lists, tuples or dicts inside of data or,
          if the data's content is longer than `max_width`
        - `2` keeps everything collapsed (all on one line)\n
        ------------------------------------------------------------------------------
        If `as_json` is set to `True`, the output will be in valid JSON format."""
        print(
            Data.to_str(data, indent, compactness, sep, max_width, as_json),
            end=end,
            flush=True,
        )

    @staticmethod
    def to_str(
        data: DataStructure,
        indent: int = 4,
        compactness: int = 1,
        max_width: int = 127,
        sep: str = ", ",
        as_json: bool = False,
    ) -> str:
        """Get nicely formatted data structure-strings.\n
        ------------------------------------------------------------------------------
        The indentation spaces-amount can be set with with `indent`.
        There are three different levels of `compactness`:
        - `0` expands everything possible
        - `1` only expands if there's other lists, tuples or dicts inside of data or,
          if the data's content is longer than `max_width`
        - `2` keeps everything collapsed (all on one line)\n
        ------------------------------------------------------------------------------
        If `as_json` is set to `True`, the output will be in valid JSON format."""

        def format_value(value: any, current_indent: int) -> str:
            if isinstance(value, dict):
                return format_dict(value, current_indent + indent)
            elif hasattr(value, "__dict__"):
                return format_dict(value.__dict__, current_indent + indent)
            elif isinstance(value, (list, tuple, set, frozenset)):
                return format_sequence(value, current_indent + indent)
            elif isinstance(value, bool):
                return str(value).lower() if as_json else str(value)
            elif isinstance(value, (int, float)):
                return "null" if as_json and (_math.isinf(value) or _math.isnan(value)) else str(value)
            elif isinstance(value, complex):
                return f"[{value.real}, {value.imag}]" if as_json else str(value)
            elif value is None:
                return "null" if as_json else "None"
            else:
                return '"' + String.escape(str(value), '"') + '"' if as_json else "'" + String.escape(str(value), "'") + "'"

        def should_expand(seq: list | tuple | dict) -> bool:
            if compactness == 0:
                return True
            if compactness == 2:
                return False
            complex_items = sum(1 for item in seq if isinstance(item, (list, tuple, dict, set, frozenset)))
            return (
                complex_items > 1
                or (complex_items == 1 and len(seq) > 1)
                or Data.chars_count(seq) + (len(seq) * len(sep)) > max_width
            )

        def format_key(k: any) -> str:
            return (
                '"' + String.escape(str(k), '"') + '"'
                if as_json
                else ("'" + String.escape(str(k), "'") + "'" if isinstance(k, str) else str(k))
            )

        def format_dict(d: dict, current_indent: int) -> str:
            if not d or compactness == 2:
                return "{" + sep.join(f"{format_key(k)}: {format_value(v, current_indent)}" for k, v in d.items()) + "}"
            if not should_expand(d.values()):
                return "{" + sep.join(f"{format_key(k)}: {format_value(v, current_indent)}" for k, v in d.items()) + "}"
            items = []
            for key, value in d.items():
                formatted_value = format_value(value, current_indent)
                items.append(f'{" " * (current_indent + indent)}{format_key(key)}: {formatted_value}')
            return "{\n" + ",\n".join(items) + f'\n{" " * current_indent}}}'

        def format_sequence(seq, current_indent: int) -> str:
            if as_json:
                seq = list(seq)
            if not seq or compactness == 2:
                return (
                    "[" + sep.join(format_value(item, current_indent) for item in seq) + "]"
                    if isinstance(seq, list)
                    else "(" + sep.join(format_value(item, current_indent) for item in seq) + ")"
                )
            if not should_expand(seq):
                return (
                    "[" + sep.join(format_value(item, current_indent) for item in seq) + "]"
                    if isinstance(seq, list)
                    else "(" + sep.join(format_value(item, current_indent) for item in seq) + ")"
                )
            items = [format_value(item, current_indent) for item in seq]
            formatted_items = ",\n".join(f'{" " * (current_indent + indent)}{item}' for item in items)
            if isinstance(seq, list):
                return "[\n" + formatted_items + f'\n{" " * current_indent}]'
            else:
                return "(\n" + formatted_items + f'\n{" " * current_indent})'

        return format_dict(data, 0) if isinstance(data, dict) else format_sequence(data, 0)

    @staticmethod
    def _is_key(data: DataStructure, path_id: str) -> bool:
        """Returns `True` if the path ID points to a key in `data` and `False` otherwise.\n
        ------------------------------------------------------------------------------------
        Input a list, tuple or dict as `data`, along with `path_id`, which is a path ID
        that was created before using `Data.get_path_id()`."""

        def check_nested(data: list | tuple | set | frozenset | dict, path: list[int]) -> bool:
            for i, idx in enumerate(path):
                if isinstance(data, dict):
                    keys = list(data.keys())
                    if i == len(path) - 1:
                        return True
                    try:
                        data = data[keys[idx]]
                    except IndexError:
                        return False
                elif isinstance(data, (list, tuple, set, frozenset)):
                    return False
                else:
                    raise TypeError(f"Unsupported type {type(data)} at path {path[:i+1]}")
            return False

        if not isinstance(data, dict):
            return False
        path = Data.__sep_path_id(path_id)
        return check_nested(data, path)

    @staticmethod
    def __sep_path_id(path_id: str) -> list[int]:
        if path_id.count(">") != 1:
            raise ValueError(f"Invalid path ID: {path_id}")
        id_part_len = int(path_id.split(">")[0])
        path_ids_str = path_id.split(">")[1]
        return [int(path_ids_str[i : i + id_part_len]) for i in range(0, len(path_ids_str), id_part_len)]



from .xx_string import String
from .xx_path import Path

import os as _os


class File:

    @staticmethod
    def rename_extension(
        file: str,
        new_extension: str,
        camel_case_filename: bool = False,
    ) -> str:
        """Rename the extension of a file.\n
        --------------------------------------------------------------------------
        If the `camel_case_filename` parameter is true, the filename will be made
        CamelCase in addition to changing the files extension."""
        directory, filename_with_ext = _os.path.split(file)
        filename = filename_with_ext.split(".")[0]
        if camel_case_filename:
            filename = String.to_camel_case(filename)
        return _os.path.join(directory, f"{filename}{new_extension}")

    @staticmethod
    def create(
        file: str,
        content: str = "",
        force: bool = False,
    ) -> str:
        """Create a file with ot without content.\n
        ------------------------------------------------------------------------
        The function will throw a `FileExistsError` if the file already exists.
        To always overwrite the file, set the `force` parameter to `True`."""
        if _os.path.exists(file) and not force:
            with open(file, "r", encoding="utf-8") as existing_file:
                existing_content = existing_file.read()
                if existing_content == content:
                    raise FileExistsError("Already created this file. (nothing changed)")
            raise FileExistsError("File already exists.")
        with open(file, "w", encoding="utf-8") as f:
            f.write(content)
        full_path = _os.path.abspath(file)
        return full_path

    @staticmethod
    def extend_or_make_path(
        file: str,
        search_in: str | list[str] = None,
        prefer_base_dir: bool = True,
        correct_paths: bool = False,
    ) -> str:
        """Tries to find the file and extend the path to be absolute and if the file was not found:\n
        Generate the absolute path to the file in the CWD or the running program's base-directory.\n
        ----------------------------------------------------------------------------------------------
        If the `file` is not found in predefined directories, it will be searched in the `search_in`
        directory/directories. If the file is still not found, it will return the path to the file in
        the base-dir per default or to the file in the CWD if `prefer_base_dir` is set to `False`.\n
        ----------------------------------------------------------------------------------------------
        If `correct_paths` is true, it is possible to have typos in the `search_in` path/s and it
        will still find the file if it is under one of those paths."""
        try:
            return Path.extend(file, search_in, raise_error=True, correct_path=correct_paths)
        except FileNotFoundError:
            return _os.path.join(Path.get(base_dir=True), file) if prefer_base_dir else _os.path.join(_os.getcwd(), file)



"""
 >>> import xulbux as xx
━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
  • CUSTOM TYPES:
     • rgba(int,int,int,float)
     • hsla(int,int,int,float)
     • hexa(str)
  • PATH OPERATIONS          xx.Path
  • FILE OPERATIONS          xx.File
  • JSON FILE OPERATIONS     xx.Json
  • SYSTEM ACTIONS           xx.System
  • MANAGE THE ENV PATH VAR  xx.EnvPath
  • CONSOLE LOG AND ACTIONS  xx.Console
  • EASY PRETTY PRINTING     xx.FormatCodes
  • WORKING WITH COLORS      xx.Color
  • DATA OPERATIONS          xx.Data
  • STRING OPERATIONS        xx.String
  • CODE STRING OPERATIONS   xx.Code
  • REGEX PATTERN TEMPLATES  xx.Regex
"""

__version__ = "1.6.5"
__author__ = "XulbuX"
__email__ = "xulbux.real@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2024 XulbuX"
__url__ = "https://github.com/XulbuX/PythonLibraryXulbuX"
__description__ = "A library which includes a lot of really helpful functions."
__all__ = [
    "_consts_",
    "xx_console",
    "xx_code",
    "xx_color",
    "xx_data",
    "xx_env_path",
    "xx_file",
    "xx_format_codes",
    "xx_json",
    "xx_path",
    "xx_regex",
    "xx_string",
    "xx_system",
]

from ._consts_ import *
from .xx_code import *
from .xx_color import *
from .xx_console import *
from .xx_data import *
from .xx_env_path import *
from .xx_file import *
from .xx_format_codes import *
from .xx_json import *
from .xx_path import *
from .xx_regex import *
from .xx_string import *
from .xx_system import *



from xulbux import Color


def test_rgba_to_hex_int_and_back():
    blue = Color.rgba_to_hex_int(0, 0, 255)
    black = Color.rgba_to_hex_int(0, 0, 0, 1)
    _blue = Color.rgba_to_hex_int(0, 0, 255, preserve_original=True)
    _black = Color.rgba_to_hex_int(0, 0, 0, 1, preserve_original=True)
    assert blue == 0x0100FF
    assert black == 0x010000FF
    assert _blue == 0x0000FF
    assert _black == 0x000000FF
    assert Color.hex_int_to_rgba(blue).values() == (0, 0, 255, None)
    assert Color.hex_int_to_rgba(black).values() == (0, 0, 0, 1.0)
    assert Color.hex_int_to_rgba(_blue).values() == (0, 0, 255, None)
    assert Color.hex_int_to_rgba(_black).values() == (0, 0, 255, None)
    assert Color.hex_int_to_rgba(blue, preserve_original=True).values() == (1, 0, 255, None)
    assert Color.hex_int_to_rgba(black, preserve_original=True).values() == (1, 0, 0, 1.0)



import re as _re


class String:

    @staticmethod
    def to_type(string: str) -> any:
        """Will convert a string to the found type."""
        string = string.strip()  # Clean up whitespace
        # BOOLEAN
        if _re.match(r"(?i)^(true|false)$", string):
            return string.lower() == "true"
        # NONE
        elif _re.match(r"(?i)^(none|null|undefined)$", string):
            return None
        # INTEGER
        elif _re.match(r"^-?\d+$", string):
            return int(string)
        # FLOAT
        elif _re.match(r"^-?\d+\.\d+$", string):
            return float(string)
        # COMPLEX
        elif _re.match(r"^(-?\d+(\.\d+)?[+-]\d+(\.\d+)?j)$", string):
            return complex(string)
        # QUOTED STRING
        elif _re.match(r'^["\'](.*)["\']$', string):
            return string[1:-1]
        # BYTES
        elif _re.match(r"^b['\"](.*)['\"]$", string):
            return bytes(string[2:-1], "utf-8")
        # LIST
        elif _re.match(r"^\[(.*)\]$", string):
            return [
                String.to_type(item.strip()) for item in _re.findall(r"(?:[^,\[\]]+|\[.*?\]|\(.*?\)|\{.*?\})+", string[1:-1])
            ]
        # TUPLE
        elif _re.match(r"^\((.*)\)$", string):
            return tuple(
                String.to_type(item.strip()) for item in _re.findall(r"(?:[^,\(\)]+|\[.*?\]|\(.*?\)|\{.*?\})+", string[1:-1])
            )
        # DICTIONARY
        elif _re.match(r"^\{(.*)\}$", string) and ":" in string:
            return {
                String.to_type(k.strip()): String.to_type(v.strip())
                for k, v in _re.findall(
                    r"((?:[^:,{}]+|\[.*?\]|\(.*?\)|\{.*?\})+)\s*:\s*((?:[^:,{}]+|\[.*?\]|\(.*?\)|\{.*?\})+)", string[1:-1]
                )
            }
        # SET
        elif _re.match(r"^\{(.*?)\}$", string):
            return {
                String.to_type(item.strip()) for item in _re.findall(r"(?:[^,{}]+|\[.*?\]|\(.*?\)|\{.*?\})+", string[1:-1])
            }
        # RETURN AS IS (str)
        return string

    @staticmethod
    def normalize_spaces(string: str, tab_spaces: int = 4) -> str:
        """Replaces all special space characters with normal spaces.
        Also replaces tab characters with `tab_spaces` spaces."""
        return (
            string.replace("\t", " " * tab_spaces)
            .replace("\u2000", " ")
            .replace("\u2001", " ")
            .replace("\u2002", " ")
            .replace("\u2003", " ")
            .replace("\u2004", " ")
            .replace("\u2005", " ")
            .replace("\u2006", " ")
            .replace("\u2007", " ")
            .replace("\u2008", " ")
            .replace("\u2009", " ")
            .replace("\u200A", " ")
        )

    @staticmethod
    def escape(string: str, str_quotes: str = '"') -> str:
        """Escapes Python's special characters (e.g. `\n`, `\t`, ...) and quotes inside the string.\n
        ----------------------------------------------------------------------------------------------
        `str_quotes` can be either `"` or `'` and should match the quotes, the string will be put
        inside of. So if your string will be `"string"`, you should pass `"` to the parameter
        `str_quotes`. That way, if the string includes the same quotes, they will be escaped."""
        string = (
            string.replace("\\", r"\\")
            .replace("\n", r"\n")
            .replace("\r", r"\r")
            .replace("\t", r"\t")
            .replace("\b", r"\b")
            .replace("\f", r"\f")
            .replace("\a", r"\a")
        )
        if str_quotes == '"':
            string = string.replace(r"\\'", "'").replace(r'"', r"\"")
        elif str_quotes == "'":
            string = string.replace(r'\\"', '"').replace(r"'", r"\'")
        return string

    @staticmethod
    def is_empty(string: str, spaces_are_empty: bool = False):
        """Returns `True` if the string is empty and `False` otherwise.\n
        -------------------------------------------------------------------------------------------
        If `spaces_are_empty` is true, it will also return `True` if the string is only spaces."""
        return (string in (None, "")) or (spaces_are_empty and isinstance(string, str) and not string.strip())

    @staticmethod
    def single_char_repeats(string: str, char: str) -> int | bool:
        """- If the string consists of only the same `char`, it returns the number of times it is present.
        - If the string doesn't consist of only the same character, it returns `False`."""
        if len(string) == len(char) * string.count(char):
            return string.count(char)
        else:
            return False

    @staticmethod
    def decompose(case_string: str, seps: str = "-_", lower_all: bool = True) -> list[str]:
        """Will decompose the string (any type of casing, also mixed) into parts."""
        return [
            (part.lower() if lower_all else part)
            for part in _re.split(rf"(?<=[a-z])(?=[A-Z])|[{_re.escape(seps)}]", case_string)
        ]

    @staticmethod
    def to_camel_case(string: str, upper: bool = True) -> str:
        """Will convert the string of any type of casing to `UpperCamelCase` or `lowerCamelCase` if `upper` is false."""
        parts = String.decompose(string)
        return ("" if upper else parts[0].lower()) + "".join(part.capitalize() for part in (parts if upper else parts[1:]))

    @staticmethod
    def to_delimited_case(string: str, delimiter: str = "_", screaming: bool = False) -> str:
        """Will convert the string of any type of casing to casing delimited by `delimiter`."""
        return delimiter.join(part.upper() if screaming else part for part in String.decompose(string))

    @staticmethod
    def get_lines(string: str, remove_empty_lines: bool = False) -> list[str]:
        """Will split the string into lines."""
        if not remove_empty_lines:
            return string.splitlines()
        lines = string.splitlines()
        if not lines:
            return []
        non_empty_lines = [line for line in lines if line.strip()]
        if not non_empty_lines:
            return []
        return non_empty_lines

    @staticmethod
    def remove_consecutive_empty_lines(string: str, max_consecutive: int = 0) -> str:
        """Will remove consecutive empty lines from the string.\n
        --------------------------------------------------------------------------------------------
        - If `max_consecutive` is `0`, it will remove all consecutive empty lines.
        - If `max_consecutive` is bigger than `0`, it will only allow `max_consecutive` consecutive
        empty lines and everything above it will be cut down to `max_consecutive` empty lines."""
        return _re.sub(r"(\n\s*){2,}", r"\1" * (max_consecutive + 1), string)

    @staticmethod
    def split_count(string: str, count: int) -> list[str]:
        """Will split the string every `count` characters."""
        return [string[i : i + count] for i in range(0, len(string), count)]



import tempfile as _tempfile
import difflib as _difflib
import shutil as _shutil
import sys as _sys
import os as _os


class Path:

    @staticmethod
    def get(cwd: bool = False, base_dir: bool = False) -> str | list[str]:
        paths = []
        if cwd:
            paths.append(_os.getcwd())
        if base_dir:
            if getattr(_sys, "frozen", False):
                base_path = _os.path.dirname(_sys.executable)
            else:
                main_module = _sys.modules["__main__"]
                if hasattr(main_module, "__file__"):
                    base_path = _os.path.dirname(_os.path.abspath(main_module.__file__))
                elif (
                    hasattr(main_module, "__spec__") and main_module.__spec__ and getattr(main_module.__spec__, "origin", None)
                ):
                    base_path = _os.path.dirname(_os.path.abspath(main_module.__spec__.origin))
                else:
                    raise RuntimeError("Can only get base directory if ran from a file.")
            paths.append(base_path)
        return paths[0] if len(paths) == 1 else paths

    @staticmethod
    def extend(path: str, search_in: str | list[str] = None, raise_error: bool = False, correct_path: bool = False) -> str:
        if path in (None, ""):
            return path

        def get_closest_match(dir: str, part: str) -> str | None:
            try:
                files_and_dirs = _os.listdir(dir)
                matches = _difflib.get_close_matches(part, files_and_dirs, n=1, cutoff=0.6)
                return matches[0] if matches else None
            except Exception:
                return None

        def find_path(start: str, parts: list[str]) -> str | None:
            current = start
            for part in parts:
                if _os.path.isfile(current):
                    return current
                closest_match = get_closest_match(current, part) if correct_path else part
                current = _os.path.join(current, closest_match) if closest_match else None
                if current is None:
                    return None
            return current if _os.path.exists(current) and current != start else None

        def expand_env_path(p: str) -> str:
            if "%" not in p:
                return p
            parts = p.split("%")
            for i in range(1, len(parts), 2):
                if parts[i].upper() in _os.environ:
                    parts[i] = _os.environ[parts[i].upper()]
            return "".join(parts)

        path = _os.path.normpath(expand_env_path(path))
        if _os.path.isabs(path):
            drive, rel_path = _os.path.splitdrive(path)
            rel_path = rel_path.lstrip(_os.sep)
            search_dirs = (drive + _os.sep) if drive else [_os.sep]
        else:
            rel_path = path.lstrip(_os.sep)
            base_dir = Path.get(base_dir=True)
            search_dirs = (
                _os.getcwd(),
                base_dir,
                _os.path.expanduser("~"),
                _tempfile.gettempdir(),
            )
        if search_in:
            search_dirs.extend([search_in] if isinstance(search_in, str) else search_in)
        path_parts = rel_path.split(_os.sep)
        for search_dir in search_dirs:
            full_path = _os.path.join(search_dir, rel_path)
            if _os.path.exists(full_path):
                return full_path
            match = find_path(search_dir, path_parts) if correct_path else None
            if match:
                return match
        if raise_error:
            raise FileNotFoundError(f"Path '{path}' not found in specified directories.")
        return _os.path.join(search_dirs[0], rel_path)

    @staticmethod
    def remove(path: str, only_content: bool = False) -> None:
        if not _os.path.exists(path):
            return None
        if not only_content:
            _shutil.rmtree(path)
        elif _os.path.isdir(path):
            for filename in _os.listdir(path):
                file_path = _os.path.join(path, filename)
                try:
                    if _os.path.isfile(file_path) or _os.path.islink(file_path):
                        _os.unlink(file_path)
                    elif _os.path.isdir(file_path):
                        _shutil.rmtree(file_path)
                except Exception as e:
                    raise Exception(f"Failed to delete {file_path}. Reason: {e}")
